"""Utilities for the analyses
=============================

"""
import re
from pathlib import Path
from textwrap import dedent

import gast as ast

from transonic.analyses import beniget
from transonic.analyses import extast

try:
    import astunparse

    dump = astunparse.dump

except ImportError:

    def dump(node):
        return ast.dump(node, indent=2)


def print_dumped(source):
    """Pretty print the AST"""
    if isinstance(source, str):
        module = extast.parse(source)
        if len(module.body) == 1:
            node = module.body[0]
        else:
            node = module
    else:
        node = source
    print(dump(node))


def print_unparsed(node):
    """Print the code corresponding to a tree or a node"""
    print(extast.unparse(node))


from transonic.analyses.capturex import CaptureX


def _fill_ast_annotations_function(function_def, ast_annotations):

    dict_node = ast_annotations.value
    for arg in function_def.args.args:
        if arg.annotation is not None:
            try:
                name = arg.arg
            except AttributeError:
                name = arg.id

            dict_node.keys.append(extast.Constant(value=name))
            dict_node.values.append(arg.annotation)


def _fill_ast_annotations_class(class_def, ast_annotations):

    dict_node = ast_annotations.value
    for node in class_def.body:
        if not isinstance(node, ast.AnnAssign):
            continue

        if node.annotation is not None:
            name = node.target.id
            dict_node.keys.append(extast.Constant(value=name))
            dict_node.values.append(node.annotation)


def get_annotations(object_def, namespace):
    """Create the annotations from a definition node"""

    # print_dump(object_def)

    ast_annotations = ast.Assign(
        targets=[extast.Name("annotations", ast.Store())],
        value=ast.Dict(keys=[], values=[]),
        type_comment=None,
    )

    if isinstance(object_def, ast.FunctionDef):
        _fill_ast_annotations_function(object_def, ast_annotations)
    elif isinstance(object_def, ast.ClassDef):
        _fill_ast_annotations_class(object_def, ast_annotations)
    else:
        raise NotImplementedError

    # print_dump(ast_annotations)

    source = extast.unparse(ast_annotations)

    try:
        del namespace["__builtins__"]
    except KeyError:
        pass
    exec(source, namespace)
    return namespace["annotations"]


def filter_code_typevars(module, duc, ancestors):
    """Create a filtered code with what is needed to create the annotations"""

    module_filtered = ast.Module()
    kept = module_filtered.body = []
    module_filtered.type_ignores = []
    suppressed = set()

    def fill_suppressed(def_):
        for user in def_.users():
            parent_in_body = ancestors.parents(user.node)[1]
            suppressed.add(parent_in_body)
            fill_suppressed(user)

    for node in module.body:
        if node in suppressed:
            continue

        if isinstance(node, ast.Import):
            if node.names[0].name in ["transonic", "numpy"]:
                kept.append(node)
            else:
                def_ = duc.chains[node.names[0]]
                fill_suppressed(def_)
            #     suppressed.add()
        elif isinstance(node, ast.ImportFrom):
            if node.module in ["transonic", "numpy"]:
                kept.append(node)

        elif isinstance(node, (ast.Assign, ast.AugAssign)):
            kept.append(node)

    return extast.unparse(module_filtered)


class AnalyseLines(ast.NodeVisitor):
    """Analyse to find the last line of a node"""

    def __init__(self, main_node):
        if isinstance(main_node, ast.Module):
            self.line_start = 1
        else:
            self.line_start = main_node.lineno
        self.line_last = self.line_start
        self.visit(main_node)

    def generic_visit(self, node):
        if hasattr(node, "lineno"):
            if node.lineno > self.line_last:
                self.line_last = node.lineno
        super().generic_visit(node)

    def get_lines(self):
        return self.line_start, self.line_last

    def get_code(self, source):
        lines = source.split("\n")

        stop = self.line_last
        nb_lines = len(lines)
        if nb_lines != stop:
            # find next not empty and not comment line
            ind = stop - 1
            next_line = ""
            while ind + 1 < nb_lines and next_line == "":
                ind += 1
                next_line = lines[ind].strip()
                if next_line.startswith("#"):
                    next_line = ""

            if any(
                next_line.startswith(character) for character in (")", "]", "}")
            ):
                stop = ind + 1

        return "\n".join(lines[self.line_start - 1 : stop])


def gather_rawcode_comments(node, code_module):
    """Get the comments in a node"""
    analysis = AnalyseLines(node)
    rawcode = dedent(analysis.get_code(code_module))
    comments = dedent(
        "\n".join(
            line for line in rawcode.split("\n") if line.strip().startswith("#")
        )
    )
    return rawcode, comments


packages_supported_by_pythran = [
    "bisect",
    "builtins",
    "cmath",
    "functools",
    "io",
    "itertools",
    "math",
    "omp",
    "operator",
    "random",
    "string",
    "time",
    "numpy",
    "numpy.fft",
    "numpy.linalg",
    "numpy.random",
    "scipy",
    "scipy.special",
]


def find_path(node: object, pathfile: str):
    """Return the path of node (instance of ast.Import or ast.ImportFrom)"""
    # FIXME find path in non local imports
    name = str()
    path = str()

    if isinstance(node, ast.ImportFrom):
        name = node.module
        if name in packages_supported_by_pythran:
            return None, None
        else:
            parent = Path(pathfile).parent
            path = parent / (str(name.replace(".", "/")) + ".py")
    else:
        if node.names[0].name in packages_supported_by_pythran:
            pass
        else:
            # TODO deal with an ast.Import
            raise NotImplementedError
    return name, path


def change_import_name(
    code_dep: str, changed_node: object, func_name: str, relative: str = None
):
    """Change the name of changed_node in code_dep by adding "__" + func + "__"
    at the beginning of the imported module, and return the modified code
    """
    mod = extast.parse(code_dep)
    for node in mod.body:
        if extast.unparse(node) == extast.unparse(changed_node):
            if isinstance(node, ast.ImportFrom):
                node.module = f"__ext__{func_name}__{node.module}"
            elif isinstance(node, ast.Import):
                node.names[0].name = f"__ext__{func_name}__{node.names[0].name}"
        if not relative:
            node.level = 0
    return extast.unparse(mod)


def filter_external_code(module: object, names: list):
    """Filter the module to keep only the necessary nodes
    needed by functions or class in the parameter names
    """
    code_dependance_annotations = ""
    lines_code = []
    for node in module.body:
        for name in names:
            if isinstance(node, ast.FunctionDef):
                if node.name == extast.unparse(name).rstrip("\n\r").strip():
                    ancestors = beniget.Ancestors()
                    ancestors.visit(module)
                    duc = beniget.DefUseChains()
                    duc.visit(module)
                    udc = beniget.UseDefChains(duc)
                    capturex = CaptureX(
                        [node],
                        module,
                        ancestors,
                        defuse_chains=duc,
                        usedef_chains=udc,
                        consider_annotations=None,
                    )
                    lines_code.append(str(extast.unparse(node)))
                    code_dependance_annotations = capturex.make_code_external()
            if isinstance(node, ast.Assign):
                if (
                    node.targets[0].id
                    == extast.unparse(name).rstrip("\n\r").strip()
                ):
                    lines_code.append(str(extast.unparse(node)))
            if isinstance(node, ast.ClassDef):
                if node.name == extast.unparse(name).rstrip("\n\r").strip():
                    lines_code.append(str(extast.unparse(node)))

    return code_dependance_annotations + "\n" + "\n".join(lines_code)


def adapt_code_dependance(func: str, codes_dependance: str, jitted_dicts: dict):
    """
    Adapt code_dependance to the call of a jitted function in a jitted function:
        - Remove the import transonic
        - Remove the jitted function statement (i.e func = jit(func))
        - Add a import statement to the jitted function
        - remove the definition of the jitted function if its on the file, or remove the import statement
    """
    special = []
    module = extast.parse(codes_dependance)
    module_body = module.body.copy()
    jitted_functions = []
    for node in module_body:
        # remove the transonic import
        if isinstance(node, ast.ImportFrom):
            if node.module == "transonic":
                module.body.remove(node)
        # remove the jitted function by jit() (i.e. func = jit(func))
        # and add the import statement for the jitted function
        elif (
            isinstance(node, ast.Assign)
            and isinstance(node.value, ast.Call)
            and node.value.func.id == "jit"
        ):
            # if assigned name different from jitted function name
            if node.targets[0].id != node.value.args[0].id:
                # change function names in jitted dict
                # The list "special" contains functions that has to be write from jitted_dicts
                # see transonic.justintime:287
                def_func = extast.unparse(jitted_dicts["functions"][func])
                spl = re.split(r"(\W+)", def_func)
                spl = [
                    node.value.args[0].id if x == node.targets[0].id else x
                    for x in spl
                ]
                st = "".join(str(e) for e in spl)
                jitted_dicts["functions"][func] = extast.parse(st)
                special.append(func)
                jitted_functions.append(node.value.args[0].id)
            else:
                jitted_functions.append(node.targets[0].id)
            module.body.remove(node)
            module.body.insert(
                0,
                [
                    extast.parse(
                        "from "
                        + node.value.args[0].id
                        + " import "
                        + node.value.args[0].id
                    )
                ],
            )
    # remove the definition:
    for node in module_body:
        # of the jitted function in the file
        if isinstance(node, ast.FunctionDef):
            if node.name in jitted_functions:
                module.body.remove(node)
        # of the jitted imported function
        if isinstance(node, ast.ImportFrom):
            for name in node.names:
                if name.name in jitted_functions:
                    node.names.remove(name)
            # remove the importFrom if no function is imported
            if not node.names:
                module.body.remove(node)
    return extast.unparse(module), jitted_dicts, special, jitted_functions


code_ext = {"function": {}, "class": {}}


def get_exterior_code(
    codes_dependance: dict,
    pathfile: str,
    previous_file_name=None,
    classes: str = None,
    relative: bool = None,
    jitted_dicts: dict = None,
):
    """Get all imported functions needed by boosted functions and methods at multiple levels,
    (i.e get functions needed by functions imported by boosted function) and add them into code_ext
    """
    special = []
    treated = []
    for func, dep in codes_dependance.items():
        if not dep:
            continue
        module_ext = extast.parse(dep)
        for node in module_ext.body:
            if not isinstance(node, (ast.ImportFrom, ast.Import)):
                continue
            # get the path of the imported module
            file_name, file_path = find_path(node, pathfile)
            # a jitted function or method needs another jitted function
            if file_name == "transonic":
                (
                    codes_dependance[func],
                    jitted_dicts,
                    spe,
                    treat,
                ) = adapt_code_dependance(
                    func, codes_dependance[func], jitted_dicts
                )
                # the "special" list signals that jitted functions has to be written
                # from jitted_dicts (see transonic/justintime.py:287)
                special = special + spe
                treated = treated + treat

    for func, dep in codes_dependance.items():
        if not dep:
            continue
        module_ext = extast.parse(dep)
        for node in module_ext.body:
            if not isinstance(node, (ast.ImportFrom, ast.Import)):
                continue
            # get the path of the imported module
            file_name, file_path = find_path(node, pathfile)
            # a jitted function or method needs another jitted function
            if not (file_name and file_name not in treated):
                continue
            new_file_name = f"__ext__{func}__{file_name}"
            # get the content of the file
            try:
                with open(str(file_path), "r") as file:
                    content = file.read()
            except:
                raise NotImplementedError(file_name + " can not be found")
            mod = extast.parse(content)
            # filter the code and add it to code_ext dict
            code_ext[classes][new_file_name] = str(
                filter_external_code(mod, node.names)
            )
            # change imported module names
            codes_dependance[func] = change_import_name(
                codes_dependance[func], node, func, relative
            )
            # recursively get the exterior codes
            if code_ext[classes][new_file_name]:
                get_exterior_code(
                    {func: code_ext[classes][new_file_name]},
                    pathfile,
                    new_file_name,
                    classes,
                )
                if previous_file_name:
                    code_ext[classes][previous_file_name] = change_import_name(
                        code_ext[classes][previous_file_name],
                        node,
                        func,
                        relative,
                    )
    return codes_dependance, code_ext, jitted_dicts, special


def extract_variable_annotations(fdef, namespace):
    variable_types = {}

    for node in fdef.body:
        if isinstance(node, ast.AnnAssign):
            name = node.target.id
            source = extast.unparse(node.annotation)
            result = eval(source, namespace)
            variable_types[name] = result

    return variable_types


def extract_returns_annotation(returns, namespace):
    if returns is None:
        return
    source = extast.unparse(returns)
    return eval(source, namespace)



"""User runtime API for the ahead-of-time compilation
=====================================================

User API
--------

.. autofunction:: boost

.. autoclass:: Transonic
   :members:
   :private-members:

Internal API
------------

.. autofunction:: _get_transonic_calling_module

.. autoclass:: CheckCompiling
   :members:
   :private-members:

"""

import inspect
import time
import subprocess
import os
import functools
import sys

from transonic.backends import backends, get_backend_name_module
from transonic.config import has_to_replace, backend_default
from transonic.log import logger
from transonic import mpi
from transonic.mpi import Path

from transonic.util import (
    get_module_name,
    has_to_compile_at_import,
    import_from_path,
    has_to_build,
    modification_date,
    is_method,
    write_if_has_to_write,
    find_module_name_from_path,
    _get_filename_from_frame,
    get_frame,
)

if mpi.nb_proc == 1:
    mpi.has_to_build = has_to_build
    mpi.modification_date = modification_date

is_transpiling = False
modules_backends = {backend_name: {} for backend_name in backends.keys()}
modules = modules_backends[backend_default]


def _get_transonic_calling_module(backend_name: str = None):
    """Get the Transonic instance corresponding to the calling module

    Parameters
    ----------

    backend_name: str

    """

    frame = get_frame(2)
    module_name = get_module_name(frame)

    if backend_name is None:
        backend_name = get_backend_name_module(module_name)

    modules = modules_backends[backend_name]

    if module_name in modules:
        ts = modules[module_name]
        if (
            ts.is_transpiling != is_transpiling
            or ts._compile_at_import_at_creation != has_to_compile_at_import()
            or (
                hasattr(ts, "path_mod")
                and ts.path_backend.exists()
                and mpi.has_to_build(ts.path_backend, ts.path_mod)
            )
        ):
            ts = Transonic(frame=frame, reuse=False, backend=backend_name)
    else:
        ts = Transonic(frame=frame, reuse=False, backend=backend_name)

    return ts


def boost(
    obj=None,
    backend: str = None,
    inline=False,
    boundscheck=True,
    wraparound=True,
    cdivision=False,
    nonecheck=True,
    nogil=False,
):
    """Decorator to declare that an object can be accelerated

    Parameters
    ----------

    obj: a function, a method or a class

    """
    if backend is not None and not isinstance(backend, str):
        raise TypeError

    ts = _get_transonic_calling_module(backend_name=backend)

    decor = ts.boost(
        inline=inline,
        nogil=nogil,
        boundscheck=boundscheck,
        wraparound=wraparound,
        cdivision=cdivision,
        nonecheck=nonecheck,
    )
    if callable(obj) or isinstance(obj, type):
        return decor(obj)
    else:
        return decor


class CheckCompiling:
    """Check if the module is being compiled and replace the module and the function"""

    def __init__(self, ts, func):
        self.has_been_replaced = False
        self.ts = ts
        self.func = func

    def __call__(self, *args, **kwargs):

        if self.has_been_replaced:
            return self.func(*args, **kwargs)

        ts = self.ts
        if ts.is_compiling and not ts.process.is_alive(raise_if_error=True):
            ts.is_compiling = False
            time.sleep(0.1)
            ts.module_backend = import_from_path(
                ts.path_extension, ts.module_backend.__name__
            )
            assert ts.backend.check_if_compiled(self.ts.module_backend)
            ts.is_compiled = True

        if not ts.is_compiling:
            self.func = getattr(ts.module_backend, self.func.__name__)
            self.has_been_replaced = True

        return self.func(*args, **kwargs)


class Transonic:
    """
    Representation of a module using ahead-of-time transonic commands

    Parameters
    ----------

    use_transonified : bool (optional, default True)

      If False, don't use the pythranized versions at run time

    frame : int (optional)

      (Internal) Frame of ``__init__`` caller.

    reuse : bool (optional, default True)

      (Internal) If True, do not recreate an instance.

    """

    def __init__(
        self, use_transonified=True, frame=None, reuse=True, backend=None
    ):

        if frame is None:
            frame = get_frame(1)

        self.module_name = module_name = get_module_name(frame)

        if backend is None:
            backend = get_backend_name_module(module_name)
        if isinstance(backend, str):
            backend = backends[backend]
        self.backend = backend
        modules = modules_backends[backend.name]

        self._compile_at_import_at_creation = has_to_compile_at_import()

        if reuse and module_name in modules:
            ts = modules[module_name]
            for key, value in ts.__dict__.items():
                self.__dict__[key] = value
            return

        self.is_transpiling = is_transpiling
        self.has_to_replace = has_to_replace

        if is_transpiling:
            self.functions = {}
            self.classes = {}
            self.signatures_func = {}
            self.is_transpiled = False
            self.is_compiled = False
            return

        self.is_compiling = False

        if not use_transonified or not has_to_replace:
            self.is_transpiled = False
            self.is_compiled = False
            return

        if "." in module_name:
            package, module_short_name = module_name.rsplit(".", 1)
            module_backend_name = package + "."
        else:
            module_short_name = module_name
            module_backend_name = ""

        module_backend_name += f"__{backend.name}__." + module_short_name

        self.path_mod = path_mod = Path(_get_filename_from_frame(frame))

        suffix = ".py"
        self.path_backend = path_backend = (
            path_mod.parent / f"__{backend.name}__" / (module_short_name + suffix)
        )

        path_ext = None

        if has_to_compile_at_import() and path_mod.exists():
            if mpi.has_to_build(path_backend, path_mod):
                if path_backend.exists():
                    time_backend = mpi.modification_date(path_backend)
                else:
                    time_backend = 0

                returncode = None
                if mpi.rank == 0:
                    print(f"Running transonic on file {path_mod}... ", end="")
                    # better to do this in another process because the file is already run...
                    os.environ["TRANSONIC_NO_MPI"] = "1"
                    returncode = subprocess.call(
                        [
                            sys.executable,
                            "-m",
                            "transonic.run",
                            "-nc",
                            str(path_mod),
                        ]
                    )
                    del os.environ["TRANSONIC_NO_MPI"]
                returncode = mpi.bcast(returncode)

                if returncode != 0:
                    raise RuntimeError(
                        f"transonic does not manage to produce the {backend.name_capitalized} "
                        f"file for {path_mod}"
                    )

                if mpi.rank == 0:
                    print("Done!")

                path_ext = path_backend.with_name(
                    backend.name_ext_from_path_backend(path_backend)
                )

                time_backend_after = mpi.modification_date(path_backend)
                # We have to touch the files to signal that they are up-to-date
                if time_backend_after == time_backend and mpi.rank == 0:
                    if not has_to_build(path_ext, path_backend):
                        path_backend.touch()
                        if path_ext.exists():
                            path_ext.touch()
                    else:
                        path_backend.touch()

        path_ext = path_ext or path_backend.with_name(
            backend.name_ext_from_path_backend(path_backend)
        )

        self.path_extension = path_ext
        if (
            has_to_compile_at_import()
            and path_mod.exists()
            and not self.path_extension.exists()
        ):
            if mpi.rank == 0:
                print(
                    f"Launching {backend.name_capitalized} to compile a new extension..."
                )
            self.is_compiling, self.process = backend.compile_extension(
                path_backend, name_ext_file=self.path_extension.name
            )
            self.is_compiled = not self.is_compiling

        self.is_transpiled = True

        if not path_ext.exists() and not self.is_compiling:
            path_ext_alt = path_backend.with_suffix(backend.suffix_extension)
            if path_ext_alt.exists():
                self.path_extension = path_ext = path_ext_alt

        self.reload_module_backend(module_backend_name)

        if not self.is_transpiled:
            logger.warning(
                f"Module {path_mod} has not been compiled for "
                f"Transonic-{backend.name_capitalized}"
            )
        else:
            self.is_compiled = backend.check_if_compiled(self.module_backend)
            if self.is_compiled:
                module = inspect.getmodule(frame)
                # module can be None if (at least) it has been run with runpy
                if module is not None:
                    if backend.name == "pythran":
                        module.__pythran__ = self.module_backend.__pythran__
                    module.__transonic__ = self.module_backend.__transonic__

            if hasattr(self.module_backend, "arguments_blocks"):
                self.arguments_blocks = getattr(
                    self.module_backend, "arguments_blocks"
                )

        modules[module_name] = self

    def reload_module_backend(self, module_backend_name=None):
        if module_backend_name is None:
            module_backend_name = self.module_backend.__name__
        if self.path_extension.exists() and not self.is_compiling:
            self.module_backend = import_from_path(
                self.path_extension, module_backend_name
            )
        elif self.path_backend.exists():
            self.module_backend = import_from_path(
                self.path_backend, module_backend_name
            )
        else:
            self.is_transpiled = False
            self.is_compiled = False

    def transonic_def(self, func):
        """Decorator used for functions

        Parameters
        ----------

        func: a function

        """
        if is_method(func):
            return self.transonic_def_method(func)

        if is_transpiling or not has_to_replace or not self.is_transpiled:
            return func

        if not hasattr(self.module_backend, func.__name__):
            self.reload_module_backend()

        try:
            func_tmp = getattr(self.module_backend, func.__name__)
        except AttributeError:
            # TODO: improve what happens in this case
            logger.warning(
                f"{self.backend.name_capitalized} file does not seem to be up-to-date:\n"
                f"{self.module_backend}\nfunc: {func.__name__}"
            )
            func_tmp = func

        if self.is_compiling:
            return functools.wraps(func)(CheckCompiling(self, func_tmp))

        return func_tmp

    def transonic_def_method(self, func):
        """Decorator used for methods

        Parameters
        ----------

        func: a function

        """

        if is_transpiling or not has_to_replace or not self.is_transpiled:
            return func

        return TransonicTemporaryMethod(func)

    def boost(self, **kwargs):
        """Universal decorator for AOT compilation

        Used for functions, methods and classes.
        """
        return self._boost_decor

    def _boost_decor(self, obj):
        """Universal decorator for AOT compilation

        Used for functions, methods and classes.
        """
        if isinstance(obj, type):
            return self.transonic_class(obj)
        else:
            return self.transonic_def(obj)

    def transonic_class(self, cls: type):
        """Decorator used for classes

        Parameters
        ----------

        cls: a class

        """
        if is_transpiling:
            return cls

        jit_methods = {
            key: value
            for key, value in cls.__dict__.items()
            if isinstance(value, TransonicTemporaryJITMethod)
        }

        if jit_methods:
            cls = jit_class(cls, jit_methods, self.backend)

        if not has_to_replace or not self.is_transpiled:
            return cls

        cls_name = cls.__name__

        for key, value in cls.__dict__.items():
            if not isinstance(value, TransonicTemporaryMethod):
                continue
            func = value.func
            func_name = func.__name__

            name_backend_func = f"__for_method__{cls_name}__{func_name}"
            name_var_code_new_method = (
                f"__code_new_method__{cls_name}__{func_name}"
            )

            if not hasattr(self.module_backend, name_backend_func):
                self.reload_module_backend()

            try:
                backend_func = getattr(self.module_backend, name_backend_func)
                code_new_method = getattr(
                    self.module_backend, name_var_code_new_method
                )
            except AttributeError:
                # TODO: improve what happens in this case
                raise RuntimeError(
                    f"{self.backend.name_capitalized} file does not seem to be up-to-date."
                )
                # setattr(cls, key, func)
            else:
                namespace = {"backend_func": backend_func}
                exec(code_new_method, namespace)
                setattr(cls, key, functools.wraps(func)(namespace["new_method"]))
        return cls

    def use_block(self, name):
        """Use the pythranized version of a code block

        Parameters
        ----------

        name : str

          The name of the block.

        """
        if not self.is_transpiled:
            raise ValueError(
                "`use_block` has to be used protected by `if ts.is_transpiled`"
            )

        if self.is_compiling and not self.process.is_alive(raise_if_error=True):
            self.is_compiling = False
            time.sleep(0.1)
            self.module_backend = import_from_path(
                self.path_extension, self.module_backend.__name__
            )
            assert self.backend.check_if_compiled(self.module_backend)
            self.is_compiled = True

        func = getattr(self.module_backend, name)
        argument_names = self.arguments_blocks[name]

        locals_caller = get_frame(1).f_locals

        arguments = [locals_caller[name] for name in argument_names]
        return func(*arguments)


class TransonicTemporaryMethod:
    """Internal temporary class for methods"""

    def __init__(self, func):
        self.func = func

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError(
            "Did you forget to decorate a class using methods decorated "
            "with transonic? Please decorate it with @boost."
        )


class TransonicTemporaryJITMethod:
    """Internal temporary class for JIT methods"""

    __transonic__ = "jit_method"

    def __init__(self, func, native, xsimd, openmp):
        self.func = func
        self.native = native
        self.xsimd = xsimd
        self.openmp = openmp

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError(
            "Did you forget to decorate a class using methods decorated "
            "with transonic? Please decorate it with @boost."
        )


def jit_class(cls, jit_methods, backend):
    """Modify the class by replacing jit methods

    1. create a Python file with @jit functions and methods
    2. import the file
    3. replace the methods

    """
    if not has_to_replace:
        return cls

    cls_name = cls.__name__
    mod_name = cls.__module__
    module = sys.modules[mod_name]

    if mod_name == "__main__":
        mod_name = find_module_name_from_path(module.__file__)

    path_jit_class = mpi.Path(backend.jit.path_class)

    # 1. create a Python file with @jit functions and methods
    python_path_dir = path_jit_class / mod_name.replace(".", os.path.sep)
    python_path = python_path_dir / (cls_name + ".py")

    if mpi.has_to_build(python_path, module.__file__):
        from transonic.justintime import _get_module_jit

        mod = _get_module_jit(backend_name=backend.name, depth_frame=5)
        if mpi.rank == 0:
            python_path = mpi.PathSeq(python_path)
            python_code = (
                mod.info_analysis["codes_dependance_classes"][cls_name] + "\n"
            )
            python_code += backend.jit.produce_code_class(cls)
            write_if_has_to_write(python_path, python_code)
            python_path = mpi.Path(python_path)
        mpi.barrier()

    # 2. import the file
    python_mod_name = path_jit_class.name + "." + mod_name + "." + cls_name
    module = import_from_path(python_path, python_mod_name)

    # 3. replace the methods
    for name_method, method in jit_methods.items():
        func = method.func
        name_new_method = f"__new_method__{cls.__name__}__{name_method}"
        new_method = getattr(module, name_new_method)
        setattr(cls, name_method, functools.wraps(func)(new_method))

    return cls



"""Logging
==========

Defines the transonic logger (variable :code:`logger`).

"""

import logging
from types import MethodType
import os

def create_logger(name, show_time=False, show_path=False):
    """Returns a logger instance using ``rich`` package if available; else
    defaults to ``logging`` standard library.

    """
    try:
        from rich.logging import RichHandler

        handler = RichHandler(show_time=show_time, show_path=show_path)
    except ImportError:
        handler = logging.StreamHandler()

    logger = logging.getLogger(name)
    logger.addHandler(handler)
    return logger


def get_logger(name):
    """Returns a logger instance using ``rich`` package if available; else
    defaults to ``logging`` standard library.

    """
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger = create_logger(name)
    return logger


logger = create_logger("transonic")


def _get_level_number(level):
    level = level.lower()
    if level == "info":
        return logging.INFO
    elif level == "debug":
        return logging.DEBUG
    elif level == "warning":
        return logging.WARNING
    else:
        raise ValueError


def set_level(self, level):
    """Set logging level"""
    if not level:
        self.setLevel(0)
        return

    level = _get_level_number(level)

    self.setLevel(level)


def get_level(self):
    return logging.getLevelName(self.getEffectiveLevel()).lower()


def is_enable_for(self, level):
    return _get_level_number(level) >= self.getEffectiveLevel()


logger.set_level = MethodType(set_level, logger)
logger.get_level = MethodType(get_level, logger)
logger.is_enable_for = MethodType(is_enable_for, logger)

if os.getenv("TRANSONIC_DEBUG"):
    level = "debug"
else:
    level = "info"

logger.set_level(level)



"""Analyses for ``@jit``
========================

"""

from transonic.analyses import extast
from transonic.analyses import compute_ancestors_chains, get_decorated_dicts
from transonic.analyses.capturex import CaptureX

from transonic.log import logger
from transonic.analyses.util import get_exterior_code


def analysis_jit(code, pathfile, backend_name):
    """Gather the informations for ``@jit`` with an ast analysis"""
    debug = logger.debug

    debug("extast.parse")
    module = extast.parse(code)

    debug("compute ancestors and chains")
    ancestors, duc, udc = compute_ancestors_chains(module)

    jitted_dicts = get_decorated_dicts(
        module, ancestors, duc, pathfile, backend_name, decorator="jit"
    )

    jitted_dicts = dict(
        functions=jitted_dicts["functions"][backend_name],
        functions_ext=jitted_dicts["functions_ext"][backend_name],
        methods=jitted_dicts["methods"][backend_name],
        classes=jitted_dicts["classes"][backend_name],
    )

    debug("compute code dependance")

    def_nodes_dict = {
        key: def_node
        for kind in ["functions"]
        for key, def_node in jitted_dicts[kind].items()
    }

    codes_dependance = {}

    # get code dependance of a decorated function from a imported module
    for func_name, (func_node, imported_module) in jitted_dicts[
        "functions_ext"
    ].items():
        capturex = CaptureX(
            (func_node,), imported_module, consider_annotations=False
        )
        codes_dependance[func_name] = capturex.make_code_external()

    # remove the decorator (jit) to compute the code dependance
    for key, def_node in def_nodes_dict.items():
        def_node.decorator_list = []

        capturex = CaptureX(
            (def_node,),
            module,
            ancestors=ancestors,
            defuse_chains=duc,
            usedef_chains=udc,
            consider_annotations=False,
        )

        codes_dependance[key] = capturex.make_code_external()

    debug(codes_dependance)

    def_nodes_dict = {}

    for (class_name, method_name), def_node in jitted_dicts["methods"].items():
        if class_name not in def_nodes_dict:
            def_nodes_dict[class_name] = []

        def_nodes_dict[class_name].append(def_node)

    codes_dependance_classes = {}

    for key, def_nodes in def_nodes_dict.items():

        for def_node in def_nodes:
            def_node.decorator_list = []

        capturex = CaptureX(
            def_nodes,
            module,
            ancestors=ancestors,
            defuse_chains=duc,
            usedef_chains=udc,
            consider_annotations=False,
        )

        codes_dependance_classes[key] = capturex.make_code_external()

    special = []
    spe = []
    if jitted_dicts["methods"]:
        codes_dependance_classes, code_ext, jitted_dicts, spe = get_exterior_code(
            codes_dependance_classes,
            pathfile,
            previous_file_name=None,
            classes="class",
            relative=False,
            jitted_dicts=jitted_dicts,
        )
    special = special + spe
    codes_dependance, code_ext, jitted_dicts, spe = get_exterior_code(
        codes_dependance,
        pathfile,
        previous_file_name=None,
        classes="function",
        relative=False,
        jitted_dicts=jitted_dicts,
    )
    special = special + spe

    return (
        jitted_dicts,
        codes_dependance,
        codes_dependance_classes,
        code_ext,
        special,
    )



"""Make the Pythran files for the classes
=========================================

"""

from tokenize import tokenize, untokenize, NAME, OP

# from token import tok_name
import inspect
from io import BytesIO

from transonic.signatures import compute_signatures_from_typeobjects

# from transonic.log import logger
from transonic.util import (
    get_source_without_decorator,
    format_str,
    make_code_from_fdef_node,
)

from .typing import base_type_formatter


def produce_code_class(cls):
    pythran_code = ""
    for key, value in list(cls.__dict__.items()):
        if hasattr(value, "__transonic__") and value.__transonic__ in (
            "trans_def_method",
            "jit_method",
        ):
            pythran_code += make_code_method_jit(cls, key)
    return pythran_code


def make_code_method_jit(cls, func_name):

    func = cls.__dict__[func_name]
    func = func.func

    new_code, attributes, name_new_func = make_new_code_method_from_objects(
        cls, func
    )

    try:
        cls_annotations = cls.__annotations__
    except AttributeError:
        cls_annotations = {}

    types_attrs = [
        cls_annotations[attr] for attr in attributes if attr in cls_annotations
    ]

    signature = inspect.signature(func)
    types_func = [param.annotation for param in signature.parameters.values()][1:]
    types_pythran = types_attrs + types_func

    transonic_signatures = "\n"

    try:
        signatures_as_lists_strings = compute_signatures_from_typeobjects(
            types_pythran, base_type_formatter
        )
    except ValueError:
        signatures_as_lists_strings = []

    for signature_as_strings in signatures_as_lists_strings:
        transonic_signatures += (
            f"# transonic def {name_new_func}("
            + ", ".join(signature_as_strings)
            + ")\n"
        )

    new_code = "from transonic import jit\n\n@jit\n" + new_code

    python_code = transonic_signatures + "\n" + new_code

    str_self_dot_attributes = ", ".join("self." + attr for attr in attributes)
    args_func = list(signature.parameters.keys())[1:]
    str_args_func = ", ".join(args_func)

    str_args_value_func = ""
    for param, value in signature.parameters.items():
        if param == "self":
            continue
        elif value.default is value.empty:
            str_args_value_func += f"{param}, "
        else:
            str_args_value_func += f"{param}={value.default}, "

    str_args_value_func = str_args_value_func.rstrip(", ")

    name_new_method = f"__new_method__{cls.__name__}__{func_name}"
    python_code += (
        f"\ndef {name_new_method}"
        f"(self, {str_args_value_func}):\n"
        f"    return {name_new_func}({str_self_dot_attributes}, {str_args_func})"
        "\n"
    )

    python_code = format_str(python_code)

    return python_code


def make_new_code_method_from_nodes(class_def, fdef):
    source = make_code_from_fdef_node(fdef)
    return make_new_code_method_from_source(source, fdef.name, class_def.name)


def make_new_code_method_from_objects(cls, func):
    source = get_source_without_decorator(func)
    return make_new_code_method_from_source(source, func.__name__, cls.__name__)


def make_new_code_method_from_source(source, func_name, cls_name):

    tokens = []
    attributes = set()

    using_self = False

    g = tokenize(BytesIO(source.encode("utf-8")).readline)
    for toknum, tokval, _, _, _ in g:
        # logger.debug((tok_name[toknum], tokval))

        if using_self == "self":
            if toknum == OP and tokval == ".":
                using_self = tokval
                continue
            elif toknum == OP and tokval in (",", ")"):
                tokens.append((NAME, "self"))
                using_self = False
            else:
                raise NotImplementedError(
                    f"self{tokval} not supported by Transonic"
                )

        if using_self == ".":
            if toknum == NAME:
                using_self = False
                tokens.append((NAME, "self_" + tokval))
                attributes.add(tokval)
                continue
            else:
                raise NotImplementedError

        if toknum == NAME and tokval == "self":
            using_self = "self"
            continue

        tokens.append((toknum, tokval))

    attributes = sorted(attributes)

    attributes_self = ["self_" + attr for attr in attributes]

    index_self = tokens.index((NAME, "self"))

    tokens_attr = []
    for ind, attr in enumerate(attributes_self):
        tokens_attr.append((NAME, attr))
        tokens_attr.append((OP, ","))

    if tokens[index_self + 1] == (OP, ","):
        del tokens[index_self + 1]

    tokens = tokens[:index_self] + tokens_attr + tokens[index_self + 1 :]

    index_func_name = tokens.index((NAME, func_name))
    name_new_func = f"__for_method__{cls_name}__{func_name}"
    tokens[index_func_name] = (NAME, name_new_func)
    # change recursive calls
    if func_name in attributes:
        attributes.remove(func_name)
        index_rec_calls = [
            index
            for index, (name, value) in enumerate(tokens)
            if value == "self_" + func_name
        ]
        # delete the occurrence of "self_" + func_name in function parameter
        del tokens[index_rec_calls[0] + 1]
        del tokens[index_rec_calls[0]]
        # consider the two deletes
        offset = -2
        # adapt all recurrence calls
        for ind in index_rec_calls[1:]:
            # adapt the index to the inserts and deletes
            ind += offset
            tokens[ind] = (tokens[ind][0], name_new_func)
            # put the attributes in parameter
            for attr in reversed(attributes):
                tokens.insert(ind + 2, (1, ","))
                tokens.insert(ind + 2, (1, "self_" + attr))
            # consider the inserts
            offset += len(attributes) * 2
    new_code = untokenize(tokens).decode("utf-8")

    return new_code, attributes, name_new_func



import sys
from distutils.core import setup

from Cython.Build import cythonize
import numpy as np

path = sys.argv.pop()
sys.argv.extend(("build_ext", "--inplace"))

setup(
    ext_modules=cythonize(path, language_level=3), include_dirs=[np.get_include()]
)



"""Base class for the Transonic backends
========================================

Internal API
------------

.. autoclass:: Backend
   :members:
   :private-members:

"""

from pathlib import Path
from textwrap import indent
from typing import Iterable, Optional

# from pprint import pprint

import transonic

from transonic.analyses import extast, analyse_aot
from transonic.log import logger
from transonic.compiler import compile_extension, ext_suffix
from transonic import mpi
from transonic.mpi import PathSeq
from transonic.signatures import compute_signatures_from_typeobjects

from transonic.util import (
    has_to_build,
    format_str,
    write_if_has_to_write,
    TypeHintRemover,
    make_hex,
)

from .base_jit import SubBackendJIT
from .for_classes import make_new_code_method_from_nodes
from .typing import TypeFormatter


class Backend:
    """Base class for the Transonic backends"""

    backend_name = "base"
    suffix_backend = ".py"
    suffix_header = None
    suffix_extension = ext_suffix
    keyword_export = "export"
    _SubBackendJIT = SubBackendJIT
    needs_compilation = True
    _TypeFormatter = TypeFormatter

    def __init__(self):
        self.name = self.backend_name
        self.name_capitalized = self.name.capitalize()
        self.type_formatter = self._TypeFormatter(self.name)
        self.jit = self._SubBackendJIT(self.name, self.type_formatter)

    def _make_code_from_fdef_node(self, fdef):
        transformed = TypeHintRemover().visit(fdef)
        # convert the AST back to source code
        code = extast.unparse(transformed)
        return format_str(code)

    def make_backend_files(self, paths_py, force=False, log_level=None):
        """Create backend files from a list of Python files"""

        if log_level is not None:
            logger.set_level(log_level)

        paths_out = []
        for path in paths_py:
            with open(path) as file:
                code = file.read()
            analyse = analyse_aot(code, path, self.name)
            path_out = self.make_backend_file(path, analyse, force=force)
            if path_out:
                paths_out.append(path_out)

        if paths_out:
            nb_files = len(paths_out)
            if nb_files == 1:
                conjug = "s"
            else:
                conjug = ""

            if self.needs_compilation:
                logger.warning(
                    f"{nb_files} files created or updated need{conjug}"
                    f" to be {self.name}ized"
                )

        return paths_out

    def make_backend_file(
        self, path_py: Path, analyse=None, force=False, log_level=None
    ):
        """Create a Python file from a Python file (if necessary)"""

        if log_level is not None:
            logger.set_level(log_level)

        path_py = Path(path_py)

        if not path_py.exists():
            raise FileNotFoundError(f"Input file {path_py} not found")

        if path_py.absolute().parent.name == f"__{self.name}__":
            logger.debug(f"skip file {path_py}")
            return None, None, None
        if not path_py.name.endswith(".py"):
            raise ValueError(
                f"transonic only processes Python file. Cannot process {path_py}"
            )

        path_dir = path_py.parent / str(f"__{self.name}__")
        path_backend = (path_dir / path_py.name).with_suffix(self.suffix_backend)

        if not has_to_build(path_backend, path_py) and not force:
            logger.warning(f"File {path_backend} already up-to-date.")
            return None, None, None

        if path_dir is None:
            return

        if not analyse:
            with open(path_py) as file:
                code = file.read()
            analyse = analyse_aot(code, path_py, self.name)

        code_backend, codes_ext, code_header = self._make_backend_code(
            path_py, analyse
        )
        if not code_backend:
            return
        logger.debug(f"code_{self.name}:\n{code_backend}")

        for file_name, code in codes_ext["function"].items():
            path_ext_file = path_dir / (file_name + ".py")
            write_if_has_to_write(
                path_ext_file, format_str(code), logger.info, force
            )

        for file_name, code in codes_ext["class"].items():
            path_ext_file = (
                path_dir.parent / f"__{self.name}__" / (file_name + ".py")
            )
            write_if_has_to_write(
                path_ext_file, format_str(code), logger.info, force
            )

        written = write_if_has_to_write(
            path_backend, code_backend, logger.info, force
        )

        if not written:
            logger.warning(f"Code in file {path_backend} already up-to-date.")
            return

        if self.suffix_header:
            path_header = (path_dir / path_py.name).with_suffix(
                self.suffix_header
            )
            write_if_has_to_write(path_header, code_header, logger.info, force)

        logger.info(f"File {path_backend} updated")

        return path_backend

    def _make_first_lines_header(self):
        return []

    def _make_beginning_code(self):
        return ""

    def _make_backend_code(self, path_py, analyse):
        """Create a backend code from a Python file"""

        boosted_dicts, code_dependance, annotations, blocks, codes_ext = analyse

        boosted_dicts = {
            key: value[self.name] for key, value in boosted_dicts.items()
        }

        lines_code = ["\n" + code_dependance + "\n"]
        lines_header = self._make_first_lines_header()
        # Deal with functions
        for fdef in boosted_dicts["functions"].values():
            signatures_func = self._make_header_1_function(fdef, annotations)
            if signatures_func:
                lines_header.extend(signatures_func)
            code_function = self._make_code_from_fdef_node(fdef)
            lines_code.append(code_function)

        # Deal with methods
        signatures, code_for_meths = self._make_code_methods(
            boosted_dicts, annotations, path_py
        )
        lines_code.extend(code_for_meths)
        if signatures:
            lines_header.extend(signatures)

        # Deal with blocks
        signatures, code_blocks = self._make_code_blocks(blocks)
        lines_code.extend(code_blocks)
        if signatures:
            lines_header.extend(signatures)

        code = "\n".join(lines_code).strip()

        if code:
            code = self._make_beginning_code() + code
            self._append_line_header_variable(lines_header, "__transonic__")
            code += f"\n\n__transonic__ = ('{transonic.__version__}',)"

        return format_str(code), codes_ext, "\n".join(lines_header).strip() + "\n"

    def _append_line_header_variable(self, lines_header, name_variable):
        pass

    def _make_code_blocks(self, blocks):
        code = []
        signatures_blocks = []
        for block in blocks:

            str_variables = ", ".join(block.signatures[0].keys())
            fdef_block = extast.gast.parse(
                f"""def {block.name}({str_variables}):pass"""
            ).body[0]

            # TODO: locals_types for blocks
            locals_types = None
            signatures_blocks.extend(
                self._make_header_from_fdef_annotations(
                    fdef_block, block.signatures, locals_types
                )
            )

            code.append(f"\ndef {block.name}({str_variables}):\n")
            code.append(indent(extast.unparse(block.ast_code), "    "))
            if block.results:
                code.append(f"    return {', '.join(block.results)}\n")

        arguments_blocks = {
            block.name: list(block.signatures[0].keys()) for block in blocks
        }

        if arguments_blocks:
            self._append_line_header_variable(
                signatures_blocks, "arguments_blocks"
            )
            code.append(f"arguments_blocks = {str(arguments_blocks)}\n")
        return signatures_blocks, code

    def _make_code_methods(self, boosted_dicts, annotations, path_py):
        meths_code = []
        header_lines = []
        for (class_name, meth_name), fdef in boosted_dicts["methods"].items():
            signatures, code_for_meth = self._make_code_method(
                class_name, fdef, meth_name, annotations, boosted_dicts
            )
            meths_code.append(code_for_meth)
            header_lines.extend(signatures)
        return header_lines, meths_code

    def _make_code_method(
        self, class_name, fdef, meth_name, annotations, boosted_dicts
    ):
        class_def = boosted_dicts["classes"][class_name]

        if class_name in annotations["classes"]:
            annotations_class = annotations["classes"][class_name]
        else:
            annotations_class = {}

        if (class_name, meth_name) in annotations["methods"]:
            annotations_meth = annotations["methods"][(class_name, meth_name)]
        else:
            annotations_meth = {}

        meth_name = fdef.name
        python_code, attributes, _ = make_new_code_method_from_nodes(
            class_def, fdef
        )

        for attr in attributes:
            if attr not in annotations_class:
                raise NotImplementedError(
                    f"self.{attr} used but {attr} not in class annotations"
                )
        types_attrs = {
            "self_" + attr: annotations_class[attr] for attr in attributes
        }
        types_pythran = {**types_attrs, **annotations_meth}

        # TODO: locals_types for methods
        locals_types = None
        signatures_method = self._make_header_from_fdef_annotations(
            extast.parse(python_code).body[0], [types_pythran], locals_types
        )

        str_self_dot_attributes = ", ".join("self." + attr for attr in attributes)
        args_func = [arg.id for arg in fdef.args.args[1:]]
        str_args_func = ", ".join(args_func)

        defaults = fdef.args.defaults
        nb_defaults = len(defaults)
        nb_args = len(fdef.args.args)
        nb_no_defaults = nb_args - nb_defaults - 1

        str_args_value_func = []
        ind_default = 0
        for ind, arg in enumerate(fdef.args.args[1:]):
            name = arg.id
            if ind < nb_no_defaults:
                str_args_value_func.append(f"{name}")
            else:
                default = extast.unparse(defaults[ind_default]).strip()
                str_args_value_func.append(f"{name}={default}")
                ind_default += 1

        str_args_value_func = ", ".join(str_args_value_func)

        if str_self_dot_attributes:
            str_args_backend_func = ", ".join(
                (str_self_dot_attributes, str_args_func)
            )
        else:
            str_args_backend_func = str_args_func

        name_var_code_new_method = f"__code_new_method__{class_name}__{meth_name}"

        self._append_line_header_variable(
            signatures_method, name_var_code_new_method
        )
        python_code += (
            f'\n{name_var_code_new_method} = """\n\n'
            f"def new_method(self, {str_args_value_func}):\n"
            f"    return backend_func({str_args_backend_func})"
            '\n\n"""\n'
        )

        return signatures_method, format_str(python_code)

    def _make_header_1_function(self, fdef, annotations):
        raise NotImplementedError

    def _make_header_from_fdef_annotations(
        self, fdef, annotations, locals_types=None, returns=None
    ):
        raise NotImplementedError

    def name_ext_from_path_backend(self, path_backend):
        """Return an extension name given the path of a Pythran file"""

        name = None
        if mpi.rank == 0:
            path_backend = PathSeq(path_backend)
            if path_backend.exists():
                with open(path_backend) as file:
                    src = file.read()
                # quick fix to recompile when the header has been changed
                for suffix in (".pythran", ".pxd"):
                    path_header = path_backend.with_suffix(suffix)
                    if path_header.exists():
                        with open(path_header) as file:
                            src += file.read()
            else:
                src = ""

            name = path_backend.stem + "_" + make_hex(src) + self.suffix_extension

        return mpi.bcast(name)

    def compile_extensions(
        self,
        paths: Iterable[Path],
        str_accelerator_flags: str,
        parallel=True,
        force=True,
    ):
        for path in paths:
            self.compile_extension(
                path,
                str_accelerator_flags=str_accelerator_flags,
                parallel=parallel,
                force=force,
            )

    def compile_extension(
        self,
        path_backend,
        name_ext_file=None,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        raise NotImplementedError


class BackendAOT(Backend):
    """Backend for ahead-of-time compilers"""

    def check_if_compiled(self, module):
        try:
            path = module.__file__
        except AttributeError:
            return True

        return not path.endswith(".py")

    def compile_extension(
        self,
        path_backend,
        name_ext_file=None,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        if name_ext_file is None:
            name_ext_file = self.name_ext_from_path_backend(path_backend)

        compiling = True
        process = compile_extension(
            path_backend,
            self.name,
            name_ext_file,
            native=native,
            xsimd=xsimd,
            openmp=openmp,
            str_accelerator_flags=str_accelerator_flags,
            parallel=parallel,
            force=force,
        )
        return compiling, process

    def _make_header_1_function(self, fdef, annotations):

        try:
            annots = annotations["__in_comments__"][fdef.name]
        except KeyError:
            annots = []

        try:
            annot = annotations["functions"][fdef.name]
        except KeyError:
            pass
        else:
            annots.append(annot)

        locals_types = annotations["__locals__"].get(fdef.name, None)
        returns = annotations["__returns__"].get(fdef.name, None)

        return self._make_header_from_fdef_annotations(
            fdef, annots, locals_types, returns
        )

    def _make_header_from_fdef_annotations(
        self, fdef, annotations, locals_types=None, returns=None
    ):
        signatures_as_lists_strings = []
        for annot in annotations:
            # print("DEBUG, annot")
            # pprint(annot)
            signatures_as_lists_strings.extend(
                compute_signatures_from_typeobjects(annot, self.type_formatter)
            )

        # print("DEBUG, signatures_as_lists_strings")
        # pprint(signatures_as_lists_strings)

        return self._make_header_from_fdef_signatures(
            fdef,
            signatures_as_lists_strings,
            locals_types=locals_types,
            returns=returns,
        )


class BackendJIT(Backend):
    """Backend for just-in-time compilers"""

    suffix_extension = ".py"
    needs_compilation = False

    def check_if_compiled(self, module):
        return True

    def _make_header_1_function(self, fdef, annotations):
        return []

    def _make_first_lines_header(self):
        return []

    def _make_header_from_fdef_annotations(
        self, fdef, annotations, locals_types=None, returns=None
    ):
        return []



"""Minimalist MPI module
========================

"""

import os
from pathlib import Path
from time import time, sleep

mpi_timeout = float(os.environ.get("TRANSONIC_MPI_TIMEOUT", "5"))


if "TRANSONIC_NO_MPI" in os.environ:
    nb_proc = 1
    rank = 0
else:
    try:
        from fluiddyn.util import mpi as _mpi
    except ImportError:
        try:
            from mpi4py import MPI
        except ImportError:
            nb_proc = 1
            rank = 0
        else:
            comm = MPI.COMM_WORLD
            nb_proc = comm.size
            rank = comm.Get_rank()
    else:
        rank = _mpi.rank
        nb_proc = _mpi.nb_proc
        if nb_proc > 1:
            comm = _mpi.comm

if nb_proc > 1:
    comm = comm.Clone()


if nb_proc == 1:

    def bcast(value, root=0):
        return value

    def barrier(timeout=None):
        pass


else:

    _tag = 3 * 7 * 9 * sum(ord(letter) for letter in "mpi transonic")

    def bcast(value, root=0, timeout=mpi_timeout, tag=_tag):
        """MPI broadcast

        Should do something similar to::

          value = comm.bcast(value, root=root)

        but with a timeout

        """

        time_comm = time()

        requests = []
        if rank == root:
            for irank in range(nb_proc):
                if irank == rank:
                    continue
                requests.append(comm.isend(value, dest=irank, tag=tag))
        else:
            requests.append(comm.irecv(source=root, tag=tag))

        for request in requests:
            while True:
                ready, received = request.test()
                if ready:
                    break
                if time() - time_comm > timeout:
                    raise TimeoutError(f"rank = {rank}, value = {value}")
                sleep(0.1)

        if rank != root:
            value = received

        # send back something small to say that we received the value
        requests = []
        if rank == root:
            for irank in range(nb_proc):
                if irank == rank:
                    continue
                requests.append(comm.irecv(source=irank, tag=tag + 100))
        else:
            requests.append(comm.isend("ok", dest=0, tag=tag + 100))

        for request in requests:
            while True:
                ready, received = request.test()
                if ready:
                    break
                if time() - time_comm > timeout:
                    raise TimeoutError(f"rank = {rank}, value = {value}")
                sleep(0.1)

        return value

    def barrier(timeout=mpi_timeout):
        if timeout is not None:
            bcast("barrier", timeout=timeout, tag=_tag + 4)
        comm.barrier()


class ShellProcessMPI:
    def __init__(self, process, root=0):

        if rank != root:
            assert process is None

        self.process = process
        self.root = root

    def is_alive(self, raise_if_error=False):
        if rank == self.root:
            is_alive = self.is_alive_root(raise_if_error=raise_if_error)
        else:
            is_alive = None
        return bcast(is_alive, self.root)

    def is_alive_root(self, raise_if_error=False):
        process = self.process
        is_alive_ = process.poll() is None

        if not is_alive_ and process.returncode:
            if process.stdout:
                print(process.stdout.read(), flush=True)
            if process.stderr:
                from transonic.log import logger

                logger.error(process.stderr.read())

            if raise_if_error:
                raise RuntimeError(
                    "Error while compiling code generated by Transonic."
                )

        return is_alive_


PathSeq = Path

if nb_proc > 1:

    class PathMPI(type(Path())):
        def exists(self):
            answer = None
            if rank == 0:
                answer = super().exists()

            ret = bcast(("exists", answer), tag=_tag + 1)

            if len(ret) != 2 and ret[0] != "exists":
                print(rank, "bug!!!!", ret, flush=1)
                raise RuntimeError

            kind, answer = ret

            return answer

        def unlink(self):
            if rank == 0:
                super().unlink()
            comm.barrier()

    Path = PathMPI

    def has_to_build(output_file: Path, input_file: Path):
        from . import util

        ret = None
        if rank == 0:
            ret = util.has_to_build(output_file, input_file)
        return bcast(ret, tag=_tag + 2)

    def modification_date(filename):
        from . import util

        ret = None
        if rank == 0:
            ret = util.modification_date(filename)
        return bcast(ret, tag=_tag + 3)


if __name__ == "__main__":
    p = Path.home() / "Dev"
    print(p.exists(), type(p))
    p = PathSeq(p)
    print(p, type(p))



"""Extended AST with CommentLine nodes
======================================

"""

from io import StringIO
from copy import deepcopy

import gast
from gast.ast3 import GAstToAst3

from transonic.analyses import beniget


class CommentLine(gast.AST):
    """New AST node representing a comment line"""

    _fields = ("s",)

    def __init__(self, s, lineno=None):
        super().__init__()
        self.s = s
        if lineno is not None:
            self.lineno = lineno

    def __deepcopy__(self, memo):
        try:
            lineno = self.lineno
        except AttributeError:
            lineno = None
        return self.__class__(deepcopy(self.s), lineno)


class GAstToAst3WithCommentLine(GAstToAst3):
    def visit_CommentLine(self, node):
        return node


def gast_to_ast(node):
    return GAstToAst3WithCommentLine().visit(node)


def parse(code, *args, **kwargs):
    """Parse a code and produce the extended AST"""
    tree = gast.parse(code, *args, **kwargs)
    CommentInserter(tree, code)
    return tree


try:
    from ast import unparse

    del unparse
except ImportError:
    # python < 3.9 (we use astunparse)

    import astunparse

    class UnparserExtended(astunparse.Unparser):
        """Unparser for extented AST"""

        def __init__(self, tree, file, with_comments=True):
            self.with_comments = with_comments
            super().__init__(tree, file=file)

        def _CommentLine(self, node):
            if self.with_comments:
                self.write(f"\n{'    '* self._indent}{node.s}")

    def unparse(tree, with_comments=True):
        """Unparse the extended AST"""

        module = type(tree).__module__
        if "gast" in module:
            tree = gast_to_ast(tree)

        v = StringIO()
        UnparserExtended(tree, file=v, with_comments=with_comments)
        return v.getvalue()


else:
    # python >= 3.9 (we use _ast._Unparser)

    import ast as _ast

    class UnparserExtended(_ast._Unparser):
        def __init__(self, *, with_comments=True, **kargs):
            self.with_comments = with_comments
            super().__init__(**kargs)

        def visit_CommentLine(self, node):
            if self.with_comments:
                self.write(f"\n{'    '* self._indent}{node.s}")

        def visit_Assign(self, node):
            try:
                node.lineno
            except AttributeError:
                node.lineno = 1
            super().visit_Assign(node)

        def visit_FunctionDef(self, node):
            try:
                node.lineno
            except AttributeError:
                node.lineno = 1
            super().visit_FunctionDef(node)

    def unparse(tree, with_comments=True):

        module = type(tree).__module__
        if "gast" in module:
            tree = gast_to_ast(tree)

        unparser = UnparserExtended(with_comments=with_comments)
        return unparser.visit(tree)


class CommentInserter(gast.NodeVisitor):
    """Insert the CommentLine nodes in an AST tree"""

    def __init__(self, tree, code):
        self.tree = tree

        self.ancestors = beniget.Ancestors()
        self.ancestors.visit(tree)

        self.code = code
        self.lines = code.split("\n")

        self.lines_comments = []
        self.lineno_comments = []
        for index, line in enumerate(self.lines):
            if line.strip().startswith("#"):
                self.lineno_comments.append(index + 1)
                self.lines_comments.append(line)

        if not self.lineno_comments:
            # nothing to do
            return

        self.last_line_comment = max(self.lineno_comments)

        self.index_comment = 0

        self.node_after = {}
        self.node_before = []

        self.previous_node = None

        self._done = False

        self.visit(tree)

        self.node_before = {
            self.lineno_comments[index]: node_before
            for index, node_before in enumerate(self.node_before)
        }

        self.modify_tree()

    def visit(self, node):

        if self._done:
            return

        if hasattr(node, "lineno") and node.lineno is not None:

            linenos = self.lineno_comments

            while node.lineno > linenos[self.index_comment]:
                self.node_after[linenos[self.index_comment]] = node
                if linenos[self.index_comment] == self.last_line_comment:
                    self._done = True
                    return
                self.index_comment += 1

            while len(self.node_after) > len(self.node_before):
                self.node_before.append(self.previous_node)

        self.previous_node = node

        super().visit(node)

    def modify_tree(self):

        # todo: debug this buggy code!

        for lineno, line in zip(self.lineno_comments, self.lines_comments):
            try:
                node_after = self.node_after[lineno]
            except KeyError:
                self.tree.body.append(CommentLine(line, lineno))
                continue
            else:
                self.insert_comment_in_parent_before_node(
                    line, lineno, node_after
                )
                continue

            try:
                node_before = self.node_before[lineno]
            except KeyError:
                pass
            else:
                if isinstance(node_before, gast.Module):
                    self.tree.body.insert(0, CommentLine(line, lineno))

    def insert_comment_in_parent_before_node(self, line, lineno, node):

        parent = self.ancestors.parent(node)
        comment = line.strip()

        for field, value in gast.iter_fields(parent):
            if isinstance(value, list):
                if node in value:
                    index = value.index(node)
                    value.insert(index, CommentLine(comment, lineno))


try:
    gast.Constant
except AttributeError:
    # gast < 0.3.0
    class Constant(gast.Str):
        def __init__(self, value):
            super().__init__(value)
            self.value = value

    class Name(gast.Name):
        def __init__(self, id="annotations", ctx=gast.Store(), annotation=None):
            super().__init__(id, ctx, annotation)

    class FunctionDef(gast.FunctionDef):
        def __init__(
            self,
            name,
            args,
            body,
            decorator_list=None,
            returns=None,
            type_comment=None,
        ):
            if decorator_list is None:
                decorator_list = []
            super().__init__(name, args, body, decorator_list, returns)


else:
    # gast >= 0.3.0
    class Constant(gast.Constant):
        def __init__(self, value):
            super().__init__(value, None)

    class Name(gast.Name):
        def __init__(self, id="annotations", ctx=gast.Store(), annotation=None):
            super().__init__(id, ctx, annotation, None)

    class FunctionDef(gast.FunctionDef):
        def __init__(
            self,
            name,
            args,
            body,
            decorator_list=None,
            returns=None,
            type_comment=None,
        ):
            if decorator_list is None:
                decorator_list = []
            super().__init__(
                name, args, body, decorator_list, returns, type_comment
            )


if __name__ == "__main__":
    code = """
# comment 0
# comment 1
if True:
# comment 1bis
    # comment 2
    # comment 3
    var_if = 1
    # comment after var_if
    if False:
        var_ifif = 1
        # comment ifif
    # comment if end
# comment 4
else:
    # comment else
    var_else = 1
# comment after else
a += 1
# comment after last
    """

    code = """
from transonic import Transonic
import numpy as np
import foo

ts = Transonic()
a = n = 1
b = np.ones(2)

if ts.is_transpiled:
    result = ts.use_block("block0")
else:
    # transonic block (
    #     A a; A1 b;
    #     float n
    # )

    # transonic block (
    #     int[:] a, b;
    #     float n
    # )

    result = a ** 2 + b.mean() ** 3 + foo.bar(n)

    """

    tree = parse(code)

    print(unparse(tree, with_comments=True))

    from transonic.analyses import compute_ancestors_chains

    compute_ancestors_chains(tree)



import gast as ast

from beniget import Ancestors, DefUseChains as DUC, UseDefChains

from beniget.beniget import Def


__all__ = ["Ancestors", "DefUseChains", "UseDefChains"]


class DefUseChains(DUC):
    def visit_List(self, node):
        if isinstance(node.ctx, ast.Load):
            dnode = self.chains.setdefault(node, Def(node))
            for elt in node.elts:
                if isinstance(elt, CommentLine):
                    continue
                self.visit(elt).add_user(dnode)
            return dnode
        # unfortunately, destructured node are marked as Load,
        # only the parent List/Tuple is marked as Store
        elif isinstance(node.ctx, ast.Store):
            return self.visit_Destructured(node)

    visit_Tuple = visit_List


# this import has to be after the definition of DefUseChains
from transonic.analyses.extast import CommentLine  # noqa: E402



"""Analyses for "if blocks"
===========================

"""

import gast as ast

from transonic.analyses.util import gather_rawcode_comments


class BlockDefinition:
    """Represent a block definition"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.__dict__.update(**kwargs)
        self.signatures = get_signatures_from_comments(self.comments)

    def __repr__(self):
        return repr(self.kwargs)


def get_block_definitions(code, module, ancestors, duc, udc):
    """Get all "if" block definitions"""
    blocks = []
    node_transonic_obj = None
    for node in module.body:
        if isinstance(node, ast.ImportFrom) and node.module == "transonic":
            for alias in node.names:
                if alias.name == "Transonic":
                    Transonic_node = alias
                    node_transonic_obj = None
                    for user in duc.chains[Transonic_node].users():
                        parent = ancestors.parent(user.node)
                        if isinstance(parent, ast.Call):
                            grandparent = ancestors.parent(parent)
                            if len(grandparent.targets) != 1:
                                continue
                            node_transonic_obj = grandparent.targets[0]
                    if node_transonic_obj is not None:
                        break

    if node_transonic_obj is None:
        return blocks

    def_ = duc.chains[node_transonic_obj]
    nodes_using_ts = [user.node for user in def_.users()]

    for user in def_.users():
        for user1 in user.users():
            if isinstance(user1.node, ast.Attribute):
                attribute = user1.node
                if attribute.attr == "is_transpiled":
                    parent = ancestors.parent(attribute)
                    if isinstance(parent, ast.If):
                        # it could be the begining of a block
                        if_node = parent
                        if len(parent.body) != 1:
                            # no it's not a block definition
                            continue
                        node = parent.body[0]
                        call = node.value
                        if isinstance(node, ast.Expr):
                            results = []
                        elif isinstance(node, ast.Assign):
                            results = [target.id for target in node.targets]
                        else:
                            # no it's not a block definition
                            continue

                        attribute = call.func
                        ts_node = attribute.value
                        if (
                            ts_node not in nodes_using_ts
                            or attribute.attr != "use_block"
                        ):
                            # no it's not a block definition
                            continue

                        try:
                            # gast >= 0.3.0 (py3.8)
                            name_block = call.args[0].value
                        except AttributeError:
                            name_block = call.args[0].s

                        rawcode, comments = gather_rawcode_comments(if_node, code)

                        # if we are here, it's a block definition
                        blocks.append(
                            BlockDefinition(
                                name=name_block,
                                results=results,
                                ast_code=if_node.orelse,
                                comments=comments,
                                rawcode=rawcode,
                            )
                        )

    return blocks


def find_index_closing_parenthesis(string: str):
    """Find the index of the closing parenthesis"""
    assert string.startswith("("), "string has to start with '('"
    stack = []
    for index, letter in enumerate(string):
        if letter == "(":
            stack.append(letter)
        elif letter == ")":
            stack.pop()
            if not stack:
                return index

    raise SyntaxError(f"Transonic syntax error for string {string}")


def get_signatures_from_comments(comments):
    """Get the blocks signatures for a block"""

    comments = comments.replace("#", "").replace("\n", "")

    signatures_tmp = [
        sig.split("->", 1)[0].strip()
        for sig in comments.split("transonic block")[1:]
    ]

    signatures = []
    for sig in signatures_tmp:
        if sig.startswith("("):
            sig = sig[1 : find_index_closing_parenthesis(sig)]
        signatures.append(sig)

    signatures_tmp = signatures
    signatures = []
    for sig_str in signatures_tmp:
        signature = {}
        type_vars_strs = [tmp.strip() for tmp in sig_str.split(";")]
        type_vars_strs = [tmp for tmp in type_vars_strs if tmp]

        for type_vars_str in type_vars_strs:
            type_as_str, vars_str = type_vars_str.strip().split(" ", 1)

            for var_str in vars_str.split(","):
                var_str = var_str.strip()
                signature[var_str] = type_as_str
        signatures.append(signature)

    return signatures



"""Backends for different accelerators
======================================

.. autosummary::
   :toctree:

   base
   base_jit
   for_classes

.. autosummary::
   :toctree:

   pythran
   cython
   numba
   py

User API
--------

.. autofunction:: set_backend_for_this_module

.. autofunction:: make_backend_files

"""

from pathlib import Path
from typing import Iterable

from transonic.config import backend_default
from transonic.util import get_module_name, get_frame

from .py import PythonBackend
from .pythran import PythranBackend
from .cython import CythonBackend
from .numba import NumbaBackend

backends = dict(
    pythran=PythranBackend(),
    cython=CythonBackend(),
    numba=NumbaBackend(),
    python=PythonBackend(),
)

backend_default_modules = {}


def set_backend_for_this_module(backend_name):
    """Programmatically set a backend for a module"""
    backend_name = backend_name.lower()
    frame = get_frame(1)
    module_name = get_module_name(frame)
    if backend_name not in backends.keys():
        raise ValueError(f"Bad backend value ({backend_name})")
    backend_default_modules[module_name] = backend_name


def get_backend_name_module(module_name):
    return backend_default_modules.get(module_name, backend_default)


def make_backend_files(
    paths: Iterable[Path],
    force=False,
    log_level=None,
    backend: str = backend_default,
):
    """Create Pythran files from a list of Python files"""
    backend = backends[backend]
    backend.make_backend_files(paths, force, log_level)



"""Cached JIT compilation
=========================

User API
--------

.. autofunction:: jit

.. autofunction:: set_compile_jit

Internal API
------------

.. autoclass:: ModuleJIT
   :members:
   :private-members:

.. autofunction:: _get_module_jit

.. autoclass:: JIT
   :members:
   :private-members:

Notes
-----

Serge talked about @jit (see https://gist.github.com/serge-sans-paille/28c86d2b33cd561ba5e50081716b2cf4)

It's indeed a good idea!

- At import time, we create one .py file per jit function.

- At run time, we create (and complete when needed) a corresponding
  .pythran file with signature(s).

  The jit decorator:

  * at the first call, get the types, create the .pythran file and call
    Pythran.

  * then, try to call the pythran function and if it fails with
    a Pythran TypeError, correct the .pythran file and recompile.

Note: During the compilation (the "warmup" of the JIT), the Python function is
used.

"""

import inspect
import itertools
import os
import sys
import time
from functools import wraps

from transonic.analyses.justintime import analysis_jit
from transonic.aheadoftime import TransonicTemporaryJITMethod
from transonic.backends import backends, get_backend_name_module
from transonic.config import has_to_replace, backend_default
from transonic.log import logger
from transonic import mpi
from transonic.util import (
    get_module_name,
    has_to_build,
    path_root,
    get_info_from_ipython,
    make_hex,
    has_to_compile_at_import,
    import_from_path,
    is_method,
    write_if_has_to_write,
    can_import_accelerator,
    format_str,
    strtobool,
    get_frame,
    _get_filename_from_frame
)

modules_backends = {backend_name: {} for backend_name in backends.keys()}
modules = modules_backends[backend_default]

_COMPILE_JIT = strtobool(os.environ.get("TRANSONIC_COMPILE_JIT", "True"))


def set_compile_jit(value):
    global _COMPILE_JIT
    _COMPILE_JIT = value


class ModuleJIT:
    """Representation of a module using jit"""

    def __init__(self, backend_name: str, frame=None):

        self.backend_name = backend_name
        if frame is None:
            frame = get_frame(1)

        self.filename = _get_filename_from_frame(frame)
        if any(self.filename.startswith(start) for start in ("<ipython-", "/tmp/ipykernel_")):
            self.is_dummy_file = True
            self._ipython_src, self.filename = get_info_from_ipython()
            self.module_name = self.filename
        else:
            self.is_dummy_file = False
            self.module_name = get_module_name(frame)
        modules_backends[backend_name][self.module_name] = self
        self.used_functions = {}
        self.jit_functions = {}

        (
            jitted_dicts,
            codes_dependance,
            codes_dependance_classes,
            code_ext,
            special,
        ) = analysis_jit(self.get_source(), self.filename, backend_name)

        self.info_analysis = {
            "jitted_dicts": jitted_dicts,
            "codes_dependance": codes_dependance,
            "codes_dependance_classes": codes_dependance_classes,
            "special": special,
        }

        self.backend = backend = backends[backend_name]
        path_jit = mpi.Path(backend.jit.path_base)
        path_jit_class = mpi.Path(backend.jit.path_class)

        # TODO: check if these files have to be written here...
        # Write exterior code for functions
        for file_name, code in code_ext["function"].items():
            path_ext = path_jit / self.module_name.replace(".", os.path.sep)
            path_ext_file = path_ext / (file_name + ".py")
            write_if_has_to_write(path_ext_file, format_str(code), logger.info)

        # Write exterior code for classes
        for file_name, code in code_ext["class"].items():
            path_ext = path_jit_class / self.module_name.replace(".", os.path.sep)
            path_ext_file = path_ext / (file_name + ".py")
            write_if_has_to_write(path_ext_file, format_str(code), logger.info)

    def get_source(self):
        if self.is_dummy_file:
            return self._ipython_src
        try:
            mod = sys.modules[self.module_name]
        except KeyError:
            with open(self.filename) as file:
                return file.read()
        else:
            return inspect.getsource(mod)


def _get_module_jit(backend_name: str = None, depth_frame: int = 2, frame=None):
    """Get the ModuleJIT instance corresponding to the calling module

    Parameters
    ----------

    depth_frame : int

      Depth of the frame to be selected

    """

    if frame is None:
        frame = get_frame(depth_frame)

    module_name = get_module_name(frame)

    if backend_name is None:
        backend_name = get_backend_name_module(module_name)

    modules = modules_backends[backend_name]

    if module_name in modules:
        return modules[module_name]
    else:
        return ModuleJIT(backend_name=backend_name, frame=frame)


def jit(func=None, backend: str = None, native=False, xsimd=False, openmp=False):
    """Decorator to record that the function has to be jit compiled

    """
    frame = get_frame(1)
    decor = JIT(frame, backend=backend, native=native, xsimd=xsimd, openmp=openmp)
    if callable(func):
        return decor(func)
    else:
        return decor


class JIT:
    """Decorator used internally by the public jit decorator
    """

    def __init__(
        self, frame, backend: str, native=False, xsimd=False, openmp=False
    ):

        self.mod = _get_module_jit(backend, frame=frame)

        self.backend = self.mod.backend
        self.native = native
        self.xsimd = xsimd
        self.openmp = openmp
        self._decorator_no_arg = False

        self.backend_func = None
        self.compiling = False
        self.process = None

    def __call__(self, func):
        if not has_to_replace:
            return func

        if is_method(func):
            return TransonicTemporaryJITMethod(
                func, self.native, self.xsimd, self.openmp
            )

        if not can_import_accelerator(self.backend.name):
            logger.warning(
                "Cannot accelerate a jitted function because "
                f"{self.backend.name_capitalized} is not importable."
            )
            return func

        func_name = func.__name__

        backend = self.backend
        mod = self.mod
        mod.jit_functions[func_name] = self
        module_name = mod.module_name

        path_jit = mpi.Path(backend.jit.path_base)
        path_backend = path_jit / module_name.replace(".", os.path.sep)

        if mpi.rank == 0:
            path_backend.mkdir(parents=True, exist_ok=True)
        mpi.barrier()

        path_backend = (path_backend / func_name).with_suffix(".py")
        if backend.suffix_header:
            path_backend_header = path_backend.with_suffix(backend.suffix_header)
        else:
            path_backend_header = False

        if path_backend.exists():
            if not mod.is_dummy_file and has_to_build(path_backend, mod.filename):
                has_to_write = True
            else:
                has_to_write = False
        else:
            has_to_write = True

        src = None

        if has_to_write:
            src, has_to_write = backend.jit.make_backend_source(
                mod.info_analysis, func, path_backend
            )

            if has_to_write and mpi.rank == 0:
                logger.debug(f"write code in file {path_backend}")
                with open(path_backend, "w") as file:
                    file.write(src)
                    file.flush()

        if src is None and mpi.rank == 0:
            with open(path_backend) as file:
                src = file.read()

        hex_src = None
        name_mod = None
        if mpi.rank == 0:
            # hash from src (to produce the extension name)
            hex_src = make_hex(src)
            name_mod = ".".join(
                path_backend.absolute()
                .relative_to(path_root)
                .with_suffix("")
                .parts
            )

        hex_src = mpi.bcast(hex_src)
        name_mod = mpi.bcast(name_mod)

        def backenize_with_new_header(arg_types="no types"):

            header_object = backend.jit.make_new_header(func, arg_types)

            header_code = backend.jit.merge_old_and_new_header(
                path_backend_header, header_object, func
            )
            backend.jit.write_new_header(
                path_backend_header, header_code, arg_types
            )

            # compute the new path of the extension
            hex_header = make_hex(header_code)
            # if mpi.nb_proc > 1:
            #     hex_header0 = mpi.bcast(hex_header)
            #     assert hex_header0 == hex_header
            name_ext_file = (
                func_name
                + "_"
                + hex_src
                + "_"
                + hex_header
                + backend.suffix_extension
            )
            self.path_extension = path_backend.with_name(name_ext_file)

            self.compiling, self.process = backend.compile_extension(
                path_backend,
                name_ext_file,
                native=self.native,
                xsimd=self.xsimd,
                openmp=self.openmp,
            )

            # for backend like numba
            if not self.compiling:
                backend_module = import_from_path(self.path_extension, name_mod)
                assert backend.check_if_compiled(backend_module)
                self.backend_func = getattr(backend_module, func_name)

        ext_files = None
        if mpi.rank == 0:
            glob_name_ext_file = (
                func_name + "_" + hex_src + "_*" + backend.suffix_extension
            )
            ext_files = list(
                mpi.PathSeq(path_backend).parent.glob(glob_name_ext_file)
            )
        ext_files = mpi.bcast(ext_files)

        if not ext_files:
            if has_to_compile_at_import() and _COMPILE_JIT:
                backenize_with_new_header()
            self.backend_func = None
        else:
            path_ext = max(ext_files, key=lambda p: p.stat().st_ctime)
            backend_module = import_from_path(path_ext, name_mod)
            self.backend_func = getattr(backend_module, func_name)

        # this is the function that will be called by the user
        @wraps(func)
        def type_collector(*args, **kwargs):

            if self.compiling:
                if not self.process.is_alive(raise_if_error=True):
                    self.compiling = False
                    time.sleep(0.1)
                    backend_module = import_from_path(
                        self.path_extension, name_mod
                    )
                    assert backend.check_if_compiled(backend_module)
                    self.backend_func = getattr(backend_module, func_name)

            try:
                return self.backend_func(*args, **kwargs)
            except TypeError as err:
                # need to compiled or recompile
                error = False
                if self.backend_func:
                    error = str(err)
                    if (
                        error.startswith("Invalid call to pythranized function `")
                        and " (reshaped)" in error
                    ):
                        logger.error(
                            "It seems that a jitted Pythran function has been called "
                            'with a "reshaped" array which is not supported by Pythran.'
                        )
                        raise
                    logger.debug(error)

            if self.compiling or not _COMPILE_JIT:
                return func(*args, **kwargs)

            if (
                self.backend_func
                and error
                and error.startswith("Invalid call to pythranized function `")
            ):
                logger.debug(error)
                logger.info(
                    f"{backend.name_capitalized} function `{func_name}` called with new types."
                )
                logger.debug(
                    "Transonic is going to recompute the function for the new types."
                )

            arg_types = [
                backend.jit.compute_typename_from_object(arg)
                for arg in itertools.chain(args, kwargs.values())
            ]

            backenize_with_new_header(arg_types)
            return func(*args, **kwargs)

        return type_collector



"""Command line
===============

Internal API
------------

.. autofunction:: main

"""

import subprocess
import sys
import logging
from pathlib import Path
import sysconfig
from time import time, sleep
from shutil import copyfile
import os

logger = logging.getLogger("transonic")
logger.setLevel(logging.INFO)

ext_suffix = sysconfig.get_config_var("EXT_SUFFIX") or ".so"


def main():
    """Minimal layer above the Pythran commandline"""

    if "-b" in sys.argv:
        if sys.argv[sys.argv.index("-b") + 1]:
            index = sys.argv.index("-b")
            backend = sys.argv[index + 1]
            del sys.argv[index]
            del sys.argv[index]
        else:
            raise ValueError("No backend is specified afert -b")
    else:
        raise ValueError("No backend is specified")

    if backend in ("python", "numba"):
        return

    compiling_name = backend.capitalize() + "izing"

    assert sys.argv[0].endswith(
        os.path.sep.join(("transonic_cl", "run_backend.py"))
    )

    args = sys.argv[1:]
    name = args[0]
    path = Path.cwd() / name

    if "-o" in args:
        index_output = args.index("-o") + 1
        name_out = args[index_output]
    else:
        name_out = Path(name).with_suffix(ext_suffix).name

    name_out_base = name_out.split(".", 1)[0]

    if "-o" in args:
        if backend == "pythran":
            name_tmp = name_out_base + ".tmp"
            args[index_output] = name_tmp
        elif backend == "cython":
            name_tmp = name_out_base + ".py"
            copyfile(name, name_tmp)
            copyfile(name.split(".", 1)[0] + ".pxd", name_out_base + ".pxd")
            name = name_tmp

        path_tmp = Path(name_tmp)
        path_out = path_tmp.with_suffix(ext_suffix)
    else:
        path_out = Path(name_out)

    name_lock = Path(name_out_base + ".lock")

    if name_lock.exists():
        print(
            f"lock file {name_lock.absolute()} present: "
            "waiting for completion of the compilation",
            flush=True,
        )
        time_out_lock = 3600  # (s) let's hope it's enough
        time_start = time()
        while name_lock.exists() and time() - time_start < time_out_lock:
            sleep(1)
        if time() - time_start >= time_out_lock:
            logger.error(f"Remove lock file {name_lock.absolute()}")
            name_lock.unlink()
            raise TimeoutError(
                f"Stop waiting for a lock file to be deleted {name_lock.absolute()}"
            )

        assert not name_lock.exists()
        sleep(1)
        if not path_out.exists():
            raise RuntimeError(
                f"After lock file were deleted, {path_out.absolute()} not created"
            )

        return

    if "-v" in args:
        # no capture_output
        stdout = stderr = None
    else:
        stdout = stderr = subprocess.PIPE

    print(f"{compiling_name} {path}", flush=True)
    if backend == "pythran":
        args.insert(0, "pythran")
        if os.getenv("TRANSONIC_DEBUG"):
            args.append("-v")
    elif backend == "cython":
        args = [sys.executable, "-m", "transonic_cl.cythonize", name]

    name_lock.touch()
    try:
        completed_process = subprocess.run(
            args, stdout=stdout, stderr=stderr, universal_newlines=True
        )
    except Exception:
        pass
    finally:
        name_lock.unlink()
    if backend == "pythran" and "-o" in args and path_tmp.exists():
        path_tmp.rename(path_out)
    elif backend == "cython":
        path_tmp.with_suffix(".c").unlink()
        path_tmp.with_suffix(".pxd").unlink()
        path_tmp.unlink()

    def log_completed_process():
        try:
            completed_process
        except NameError:
            pass
        else:
            if completed_process.stdout:
                print(
                    f"{backend.capitalize()} stdout:\n{completed_process.stdout}"
                )
            if completed_process.stderr:
                logger.error(
                    f"{backend.capitalize()} stderr:\n{completed_process.stderr}"
                )

    if path_out.exists():
        print(f"File {path_out.absolute()} created by {backend}", flush=True)
        if os.getenv("TRANSONIC_DEBUG"):
            log_completed_process()
    else:
        logger.error(
            f"Error! File {path_out.absolute()} has not been created by {backend}"
        )
        log_completed_process()
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Utilities to replace strings by objects
==========================================

"""

import gast as ast

from transonic.analyses import extast
from transonic.analyses.capturex import CaptureX


def find_last_def_node(variable, module):
    for node in module.body[::-1]:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if target.id == variable:
                    return node
    raise RuntimeError


def find_def_code(variables: set, module: dict, ancestors, udc, duc):

    nodes_def_vars = [
        find_last_def_node(variable, module) for variable in variables
    ]
    nodes_def_vars.sort(key=lambda x: x.lineno)

    capturex = CaptureX(
        list(nodes_def_vars),
        module,
        ancestors,
        defuse_chains=duc,
        usedef_chains=udc,
    )

    lines_ext = []
    for node in capturex.external:
        lines_ext.append(extast.unparse(node).strip())

    for node in nodes_def_vars:
        line = extast.unparse(node).strip()
        if line not in lines_ext:
            lines_ext.append(line)

    return "\n".join(lines_ext)


def create_objects_from_names(names, module, ancestors, udc, duc):
    """create a namespace for the variables defined at the module level"""
    code_def_var = find_def_code(names, module, ancestors, udc, duc,)
    namespace = {}
    exec(code_def_var, namespace)
    return namespace


def replace_strings_by_objects(signatures: dict, module, ancestors, udc, duc):
    """Replace strings in signatures by objects defined in the code"""

    types_NameError = set()
    for signature in signatures:
        for var_str, type_as_str in signature.items():
            try:
                signature[var_str] = eval(type_as_str)
            except NameError:
                types_NameError.add(type_as_str)
            except (SyntaxError, TypeError):
                pass

    namespace = create_objects_from_names(
        types_NameError, module, ancestors, udc, duc
    )

    for signature in signatures:
        for var_str, type_as_str in signature.items():
            if type_as_str in types_NameError:
                signature[var_str] = namespace[type_as_str]



"""Display progress on the console
==================================

Relies on the rich package if it is installed. If not, fall back to simple
logging messages.

"""
from transonic.log import logger


try:
    import rich
except ImportError:

    def track(sequence, *args, **kwargs):
        return sequence

    def _add_task(description, *args, **kwargs):
        logger.info(description)

    def Progress(*args, **kwargs):
        """Mock all methods of the Progress class except add_task."""
        from unittest.mock import MagicMock

        return MagicMock(name="Progress", add_task=_add_task)

    logger.debug("Install rich for tracking progress.")
else:
    logger.debug("Using rich for tracking progress.")
    from rich.progress import track, Progress



"""Pythran Backend
==================

Internal API
------------

.. autoclass:: PythranBackend
   :members:
   :private-members:

"""

from .base import BackendAOT


class PythranBackend(BackendAOT):
    """Main class for the Pythran backend"""

    backend_name = "pythran"
    suffix_header = ".pythran"

    def check_if_compiled(self, module):
        return hasattr(module, "__pythran__")

    def _append_line_header_variable(self, lines_header, name_variable):
        lines_header.append(f"export {name_variable}\n")

    def _make_header_from_fdef_signatures(
        self, fdef, signatures_as_lists_strings, locals_types=None, returns=None
    ):
        nb_defaults = len(fdef.args.defaults)
        if nb_defaults:
            new_signatures = []
            for signature_as_strings in signatures_as_lists_strings:
                for nb_args_not_given in range(1, nb_defaults + 1):
                    new_signatures.append(
                        signature_as_strings[:-nb_args_not_given]
                    )
            signatures_as_lists_strings.extend(new_signatures)

        signatures_func = set(
            f"export {fdef.name}({', '.join(signature_as_strings)})"
            for signature_as_strings in signatures_as_lists_strings
        )

        if not fdef.args.args and not signatures_func:
            signatures_func.add(f"export {fdef.name}()")

        signatures_func = sorted(signatures_func)
        if signatures_func:
            signatures_func[-1] = signatures_func[-1] + "\n"
        return signatures_func



"""Create Pythran signatures from type hints
============================================

User API
--------

.. autoclass:: Type
   :members:

.. autoclass:: NDim
   :members:

.. autoclass:: Array
   :members:
   :private-members:

.. autoclass:: List
   :members:
   :private-members:

.. autoclass:: Tuple
   :members:
   :private-members:

.. autoclass:: Dict
   :members:
   :private-members:

.. autoclass:: Set
   :members:
   :private-members:

.. autoclass:: Union
   :members:
   :private-members:

.. autofunction:: str2type

.. autofunction:: typeof

.. autofunction:: const

Internal API
------------

.. autoclass:: TemplateVar
   :members:
   :private-members:

.. autoclass:: ArrayMeta
   :members:
   :private-members:

.. autoclass:: ListMeta
   :members:
   :private-members:

.. autoclass:: DictMeta
   :members:
   :private-members:

.. autofunction:: format_type_as_backend_type

.. autoclass:: ConstType
   :members:
   :private-members:

"""
import re
from enum import Enum, auto
import itertools

import numpy as np

from transonic.util import get_name_calling_module

names_template_variables = {}


class FusedType:
    def is_fused_type(self):
        raise NotImplementedError

    def get_all_formatted_backend_types(self, type_formatter):

        template_params = self.get_template_parameters()
        values_template_parameters = {
            param.__name__: param.values for param in template_params
        }
        names = tuple(values_template_parameters.keys())
        formatted_types = []
        for set_types in itertools.product(*values_template_parameters.values()):
            template_variables = dict(zip(names, set_types))
            formatted_types.append(
                format_type_as_backend_type(
                    self, type_formatter, **template_variables
                )
            )
        return formatted_types


class TemplateVar:
    """Base class for template variables

    >>> T = TemplateVar("T")
    >>> T = TemplateVar("T", int, float)

    >>> T = TemplateVar()
    Traceback (most recent call last):
        ...
    ValueError

    >>> T = TemplateVar(1)
    Traceback (most recent call last):
        ...
    TypeError: (1,) [False]
    """

    _type_values = type
    _letter = "T"

    def get_template_parameters(self):
        return (self,)

    def __init__(self, *args, name_calling_module=None):

        if not args:
            raise ValueError

        if name_calling_module is None:
            name_calling_module = get_name_calling_module()

        if name_calling_module not in names_template_variables:
            names_template_variables[name_calling_module] = {}

        names_variables = names_template_variables[name_calling_module]

        if type(self) not in names_variables:
            names_variables[type(self)] = set()

        names_already_used = names_variables[type(self)]

        if self._is_correct_for_name(args[0]):
            self.__name__ = args[0]
            args = args[1:]
        else:
            index_var = len(names_already_used)

            while self._letter + str(index_var) in names_already_used:
                index_var += 1

            self.__name__ = self._letter + str(index_var)

        self.values = args

        names_already_used.add(self.__name__)

        self._check_type_values()

    def _is_correct_for_name(self, arg):
        return isinstance(arg, str)

    def _check_type_values(self):
        if not all(isinstance(value, self._type_values) for value in self.values):
            raise TypeError(
                f"{self.values} "
                f"{[isinstance(value, self._type_values) for value in self.values]}"
            )

    def has_multiple_values(self):
        return len(self.values) > 1


class Type(TemplateVar, FusedType):
    """Template variable representing the dtype of an array.

    As a user, it is useful only for fused types.

    >>> Type(int, float)
    Type(int, float)

    """

    def __repr__(self):
        repr_values = []
        for value in self.values:
            if hasattr(value, "__name__"):
                repr_values.append(value.__name__)
            else:
                repr_values.append(repr(value))
        return f"Type({', '.join(repr_values)})"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):

        dtype = None

        for key, value in kwargs.items():
            if key == self.__name__:
                dtype = value
                break

        if dtype is None:
            raise ValueError

        return dtype.__name__

    def is_fused_type(self):
        return len(self.values) > 1

    def short_repr(self):
        long_repr = repr(self)
        replaced_by = {"(": "I", ")": "I", ", ": "_"}
        for replaced, replacer in replaced_by.items():
            long_repr = long_repr.replace(replaced, replacer)
        return long_repr


class NDim(TemplateVar):
    """Template variable representing the number of dimension of an array.

    As a user, it is useful only for fused types.

    >>> N = NDim(1, 2)
    >>> N1 = N + 1

    """

    _type_values = int
    _letter = "N"

    def __init__(self, *args, shift=0, name_calling_module=None):

        if name_calling_module is None:
            name_calling_module = get_name_calling_module()

        super().__init__(*args, name_calling_module=name_calling_module)
        self.shift = shift

    def __repr__(self):
        if len(self.values) == 1:
            name = f'"{self.values[0]}d"'
        else:
            name = f"NDim({', '.join(repr(v) for v in self.values)})"

        if self.shift == 0:
            return name
        elif self.shift < 0:
            return name + f" - {-self.shift}"
        elif self.shift > 0:
            return name + f" + {self.shift}"
        else:
            raise RuntimeError

    def __add__(self, number):
        name_calling_module = get_name_calling_module()
        return type(self)(
            self.__name__,
            *self.values,
            shift=number,
            name_calling_module=name_calling_module,
        )

    def __sub__(self, number):
        name_calling_module = get_name_calling_module()
        return type(self)(
            self.__name__,
            *self.values,
            shift=-number,
            name_calling_module=name_calling_module,
        )

    def short_repr(self):
        long_repr = repr(self)
        replaced_by = {
            '"': "",
            "(": "I",
            ")": "I",
            " - ": "m",
            " + ": "p",
            ", ": "_",
        }
        for replaced, replacer in replaced_by.items():
            long_repr = long_repr.replace(replaced, replacer)
        return long_repr


class UnionVar(TemplateVar):
    """TemplateVar used for the Union type"""

    _type_values = (type, type(None))
    _letter = "U"


class Meta(type, FusedType):
    """Type of the Transonic types (used to create metaclasses)"""

    def __call__(cls, *args, **kwargs):
        raise RuntimeError("Transonic types are not meant to be instantiated")

    def is_fused_type(self):
        template_parameters = self.get_template_parameters()
        for template_parameter in template_parameters:
            if hasattr(template_parameter, "is_fused_type"):
                if template_parameter.is_fused_type():
                    return True
            if hasattr(template_parameter, "has_multiple_values"):
                if template_parameter.has_multiple_values():
                    return True
        return False


class MemLayout(Enum):
    C = auto()
    F = auto()
    C_or_F = auto()
    strided = auto()

    def __repr__(self):
        return f'"{self.name}"'


def str2shape(str_shape):
    assert str_shape.startswith("[") and str_shape.endswith("]")
    str_shape = str_shape.replace(" ", "")
    if str_shape == "[]":
        return (None,)
    n = str_shape.count("]")
    if n > 1:
        return (None,) * n
    shape = []
    for symbol in str_shape[1:-1].split(","):
        if symbol == ":":
            value = None
        elif symbol == "":
            continue
        else:
            value = int(symbol)
        shape.append(value)
    return tuple(shape)


def shape2str(shape):
    symbols = [":" if value is None else str(value) for value in shape]
    tmp = ",".join(symbols)
    return f'"[{tmp}]"'


class ArrayMeta(Meta):
    """Metaclass for the Array class"""

    def __getitem__(self, parameters):

        if not isinstance(parameters, tuple):
            parameters = (parameters,)

        dtype = None
        ndim = None
        memview = False
        mem_layout = MemLayout.C_or_F
        shape = None
        positive_indices = False
        params_filtered = []
        for param in parameters:
            if param is None:
                continue
            if isinstance(param, (Type, type, np.dtype)):
                if dtype is not None:
                    raise ValueError(
                        "Array should be defined with only one variable defining "
                        "the types. For more than one type, use "
                        "for example Type(float, int)"
                    )
                if isinstance(param, np.dtype):
                    param = param.type

                dtype = param

            if isinstance(param, NDim):
                if ndim is not None:
                    raise ValueError(
                        "Array should be defined with only "
                        "one NDim. For more than one dimension, use "
                        "for example NDim(2, 3)."
                    )
                ndim = param

            if (
                isinstance(param, str)
                and param[-1] == "d"
                and param[:-1].isnumeric()
            ):
                try:
                    tmp = int(param[:-1])
                except ValueError:
                    pass
                else:
                    if ndim is not None:
                        raise ValueError(
                            "Array should be defined with only "
                            "one string fixing the number of dimension. "
                            "Use for example NDim(2, 3)."
                        )
                    param = ndim = NDim(
                        tmp, name_calling_module=get_name_calling_module()
                    )

            if isinstance(param, str):
                param = param.strip()
                if param == "memview":
                    memview = True
                    continue
                if param == "positive_indices":
                    positive_indices = True
                    continue
                if param.startswith("[") and param.endswith("]"):
                    shape = str2shape(param)
                    continue
                try:
                    mem_layout = MemLayout[param]
                    continue
                except KeyError:
                    pass
                raise ValueError(f"{param} cannot be interpretted...")

            params_filtered.append(param)

        if shape is not None:
            if ndim is None:
                ndim = NDim(
                    len(shape), name_calling_module=get_name_calling_module()
                )
                params_filtered.append(ndim)
            elif ndim != len(shape):
                raise ValueError("ndim != len(shape)")
            if not any(shape):
                shape = None

        if dtype is None:
            raise ValueError("No way to determine the dtype of the array")

        if ndim is None:
            raise ValueError("No way to determine the ndim of the array")

        parameters = {p.__name__: p for p in params_filtered}

        assert isinstance(ndim, NDim)

        if hasattr(dtype, "short_repr"):
            dtype_name = dtype.short_repr()
        else:
            dtype_name = dtype.__name__

        return type(
            f"Array_{dtype_name}_{ndim.short_repr()}",
            (Array,),
            {
                "dtype": dtype,
                "ndim": ndim,
                "parameters": parameters,
                "memview": memview,
                "mem_layout": mem_layout,
                "shape": shape,
                "positive_indices": positive_indices,
            },
        )

    def get_parameters(self):
        return getattr(self, "parameters", dict())

    def get_template_parameters(self):
        return tuple(
            param
            for param in self.get_parameters().values()
            if isinstance(param, TemplateVar)
        )

    def __repr__(self):
        if not hasattr(self, "parameters"):
            return super().__repr__()

        if self.shape is not None:
            parameters = [
                param
                for param in self.parameters.values()
                if not isinstance(param, NDim)
            ]
        else:
            parameters = self.parameters.values()

        strings = []
        for p in parameters:
            if isinstance(p, type):
                string = p.__name__
            else:
                string = repr(p)
            strings.append(string)

        if self.shape is not None:
            strings.append(shape2str(self.shape))

        if self.memview:
            strings.append('"memview"')

        if self.mem_layout is not MemLayout.C_or_F:
            strings.append(repr(self.mem_layout))

        if self.positive_indices:
            strings.append('"positive_indices"')

        return f"Array[{', '.join(strings)}]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):

        dtype = ndim = None

        for var in self.parameters.values():
            if isinstance(var, Type) and var.values:
                dtype = var.values[0]
            elif isinstance(var, NDim) and var.values:
                ndim = var.values[0]
            elif isinstance(var, type):
                dtype = var

        for key, value in kwargs.items():
            try:
                template_var = self.parameters[key]
            except KeyError:
                continue

            if isinstance(template_var, Type):
                dtype = value
            elif isinstance(template_var, NDim):
                ndim = value + template_var.shift
            else:
                raise ValueError

        if dtype is None or ndim is None:
            raise ValueError

        memview = kwargs.get("memview", self.memview)
        return backend_type_formatter.make_array_code(
            dtype,
            ndim,
            self.shape,
            memview,
            self.mem_layout,
            self.positive_indices,
        )


class Array(metaclass=ArrayMeta):
    """Represent a Numpy array.

    >>> Array[int, "2d"]
    Array[int, "2d"]

    >>> Array[int, "2d", "C"]
    Array[int, "2d", "C"]

    >>> Array[int, "2d", "F"]
    Array[int, "2d", "F"]

    >>> Array[int, "2d", "strided"]
    Array[int, "2d", "strided"]

    Fused types:

    >>> Array[Type(int, float), "1d"]
    Array[Type(int, float), "1d"]

    >>> Array[float, NDim(2, 3)]
    Array[float, NDim(2, 3)]

    >>> Array[int, "1d", "C", "positive_indices"]
    Array[int, "1d", "C", "positive_indices"]

    """


class UnionMeta(Meta):
    """Metaclass for the Union class"""

    def __getitem__(self, types):

        types_in = types
        if not isinstance(types_in, tuple):
            types_in = (types_in,)

        types = []
        for type_ in types_in:
            if isinstance(type_, str):
                type_ = str2type(type_)
            types.append(type_)
        types = tuple(types)

        name_calling_module = get_name_calling_module()
        template_var = UnionVar(*types, name_calling_module=name_calling_module)

        short_repr = []
        for value in types:
            if hasattr(value, "short_repr"):
                short_repr.append(value.short_repr())
            elif hasattr(value, "__name__"):
                short_repr.append(value.__name__)
            else:
                short_repr.append(repr(value))

        return type(
            f"Union{'_'.join(short_repr)}",
            (Union,),
            {"types": types, "template_var": template_var},
        )

    def get_template_parameters(self):
        template_params = []
        for type_ in self.types:
            if hasattr(type_, "get_template_parameters"):
                template_params.extend(type_.get_template_parameters())
        template_params.append(self.template_var)
        return tuple(template_params)

    def __repr__(self):
        strings = []
        if not hasattr(self, "types"):
            return super().__repr__()

        for p in self.types:
            if isinstance(p, Meta):
                string = repr(p)
            elif isinstance(p, type):
                string = p.__name__
            else:
                string = repr(p)
            strings.append(string)
        return "Union[" + ", ".join(strings) + "]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        type_ = kwargs.pop(self.template_var.__name__)
        return format_type_as_backend_type(
            type_, backend_type_formatter, **kwargs
        )

    def short_repr(self):
        return self.__name__


class Union(metaclass=UnionMeta):
    """Similar to typing.Union

    >>> Union[float, Array[int, "1d"]]
    Union[float, Array[int, "1d"]]

    """


class ListMeta(Meta):
    """Metaclass for the List class"""

    def __getitem__(self, type_elem):
        if isinstance(type_elem, str):
            type_elem = str2type(type_elem)
        return type("ListBis", (List,), {"type_elem": type_elem})

    def get_template_parameters(self):
        if hasattr(self.type_elem, "get_template_parameters"):
            return self.type_elem.get_template_parameters()
        return tuple()

    def __repr__(self):
        if not hasattr(self, "type_elem"):
            return super().__repr__()
        if isinstance(self.type_elem, Meta):
            string = repr(self.type_elem)
        elif isinstance(self.type_elem, type):
            string = self.type_elem.__name__
        else:
            string = repr(self.type_elem)
        return f"List[{string}]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        return backend_type_formatter.make_list_code(self.type_elem, **kwargs)


class List(metaclass=ListMeta):
    """Similar to typing.List

    >>> List[List[int]]
    List[List[int]]

    """


class DictMeta(Meta):
    """Metaclass for the Dict class"""

    def __getitem__(self, types):
        type_keys, type_values = types
        if isinstance(type_keys, str):
            type_keys = str2type(type_keys)
        if isinstance(type_values, str):
            type_values = str2type(type_values)
        return type(
            "DictBis",
            (Dict,),
            {"type_keys": type_keys, "type_values": type_values},
        )

    def get_template_parameters(self):
        template_params = []
        if hasattr(self.type_keys, "get_template_parameters"):
            template_params.extend(self.type_keys.get_template_parameters())
        if hasattr(self.type_values, "get_template_parameters"):
            template_params.extend(self.type_values.get_template_parameters())
        return template_params

    def __repr__(self):
        if not hasattr(self, "type_keys"):
            return super().__repr__()
        if isinstance(self.type_keys, type):
            key = self.type_keys.__name__
        else:
            key = repr(self.type_keys)
        if isinstance(self.type_values, type):
            value = self.type_values.__name__
        else:
            value = repr(self.type_values)
        return f"Dict[{key}, {value}]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        return backend_type_formatter.make_dict_code(
            self.type_keys, self.type_values, **kwargs
        )


class Dict(metaclass=DictMeta):
    """Similar to typing.Dict

    >>> Dict[str, int]
    Dict[str, int]

    """


class SetMeta(Meta):
    """Metaclass for the Set class"""

    def __getitem__(self, type_keys):
        if isinstance(type_keys, str):
            type_keys = str2type(type_keys)
        return type("SetBis", (Set,), {"type_keys": type_keys})

    def get_template_parameters(self):
        if hasattr(self.type_keys, "get_template_parameters"):
            return self.type_keys.get_template_parameters()
        else:
            return tuple()

    def __repr__(self):
        if not hasattr(self, "type_keys"):
            return super().__repr__()
        if isinstance(self.type_keys, type):
            key = self.type_keys.__name__
        else:
            key = repr(self.type_keys)
        return f"Set[{key}]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        return backend_type_formatter.make_set_code(self.type_keys, **kwargs)


class Set(metaclass=SetMeta):
    """Similar to typing.Set

    >>> Set[str]
    Set[str]

    """


class TupleMeta(Meta):
    """Metaclass for the Tuple class"""

    def __getitem__(self, types):

        if not isinstance(types, tuple):
            types = (types,)

        trans_types = []
        for type_in in types:
            if isinstance(type_in, str):
                type_in(str2type(type_in))
            trans_types.append(type_in)

        return type("TupleBis", (Tuple,), {"types": trans_types})

    def get_template_parameters(self):
        template_params = []
        for type_ in self.types:
            if hasattr(type_, "get_template_parameters"):
                template_params.extend(type_.get_template_parameters())
        return tuple(template_params)

    def __repr__(self):
        if not hasattr(self, "types"):
            return super().__repr__()
        strings = []
        for type_ in self.types:
            if isinstance(type_, Meta):
                name = repr(type_)
            elif isinstance(type_, type):
                name = type_.__name__
            else:
                name = repr(type_)
            strings.append(name)
        return f"Tuple[{', '.join(strings)}]"

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        return backend_type_formatter.make_tuple_code(self.types, **kwargs)


class Tuple(metaclass=TupleMeta):
    """Similar to typing.Tuple

    >>> Tuple[int, Array[int, "2d"]]
    Tuple[int, Array[int, "2d"]]

    """


class OptionalMeta(Meta):
    def __getitem__(self, type_):
        return Union[type_, None]


class Optional(metaclass=OptionalMeta):
    """Similar to typing.Optional

    >>> Optional[int]
    Union[int, None]

    """


def format_type_as_backend_type(type_, backend_type_formatter, **kwargs):
    """Format a Transonic type as a backend (Pythran, Cython, ...) type"""
    if type_ is None:
        # None has a special meaning for typing...
        return "None"

    if isinstance(type_, str):
        type_ = str2type(type_)

    if hasattr(type_, "format_as_backend_type"):
        backend_type = type_.format_as_backend_type(
            backend_type_formatter, **kwargs
        )
    elif hasattr(type_, "__name__"):
        backend_type = type_.__name__
    else:
        print("type_", type_, type(type_))
        raise RuntimeError(f"type_: {type_}")

    assert backend_type is not None

    return backend_type_formatter.normalize_type_name(backend_type)


def str2type(str_type):
    """Compute a Transonic type from a string

    >>> str2type("int[:,:]")
    Array[int, "2d"]

    >>> str2type("int or float[]")
    Union[int, Array[float, "1d"]]

    >>> str2type("(int, float[:, :])")
    Tuple[int, Array[float, "2d"]]
    """
    str_type = str_type.strip()

    if " or " in str_type:
        subtypes = str_type.split(" or ")
        return Union[tuple(str2type(subtype) for subtype in subtypes)]

    try:
        return eval(str_type)
    except (TypeError, SyntaxError, NameError):
        # not a simple type
        pass

    if "[" not in str_type:
        # could be a numpy type
        try:
            if not str_type.startswith("np."):
                dtype = "np." + str_type
            else:
                dtype = str_type
            return eval(dtype, {"np": np})
        except (TypeError, SyntaxError, AttributeError):
            pass

    if str_type.startswith("(") and str_type.endswith(")"):
        re_comma = re.compile(r",(?![^\[]*\])(?![^\(]*\))")
        return Tuple[
            tuple(
                str2type(word) for word in re_comma.split(str_type[1:-1]) if word
            )
        ]

    words = [word for word in str_type.split(" ") if word]

    if words[-1] == "list":
        return List[" ".join(words[:-1])]

    if words[-1] == "dict":
        if len(words) != 3:
            raise NotImplementedError(f"words: {words}")
        key = words[0][:-1]
        value = words[1]
        return Dict[key, value]

    if words[-1] == "set":
        if len(words) != 2:
            raise NotImplementedError(f"words: {words}")
        key = words[0]
        return Set[key]

    # str_type should be of the form "int[]"
    if "[" not in str_type:
        raise ValueError(f"Can't determine the Transonic type from '{str_type}'")

    dtype, str_shape = str_type.split("[", 1)
    dtype = dtype.strip()
    if not dtype.startswith("np.") and dtype not in ("int", "float"):
        dtype = "np." + dtype
    str_shape = "[" + str_shape
    dtype = eval(dtype, {"np": np})
    return Array[dtype, str_shape]


_simple_types = (int, float, complex, str)


def typeof(obj):
    """Compute the Transonic type corresponding to a Python object

    Supports:

    - simple Python types (int, float, complex, str)
    - homogeneous list, dict and set
    - tuple
    - numpy scalars
    - numpy arrays

    """
    if isinstance(obj, _simple_types):
        return type(obj)

    if isinstance(obj, tuple):
        return Tuple[tuple(typeof(elem) for elem in obj)]

    if isinstance(obj, (list, dict, set)) and not obj:
        raise ValueError(
            f"Cannot determine the full type of an empty {type(obj)}"
        )

    if isinstance(obj, list):
        type_elem = type(obj[0])
        if not all(isinstance(elem, type_elem) for elem in obj):
            raise ValueError(f"The list {obj} is not homogeneous in type")

        return List[typeof(obj[0])]

    if isinstance(obj, (dict, set)):
        key = next(iter(obj))
        type_key = type(key)
        if not all(isinstance(key, type_key) for key in obj):
            raise ValueError("The dict {obj} is not homogeneous in type")

        if isinstance(obj, dict):
            value = next(iter(obj.values()))
            type_value = type(value)
            if not all(isinstance(value, type_value) for value in obj.values()):
                raise ValueError("The dict {obj} is not homogeneous in type")
            return Dict[typeof(key), typeof(value)]
        else:
            return Set[typeof(key)]

    # TODO: Tuple
    if isinstance(obj, tuple):
        raise NotImplementedError

    if isinstance(obj, np.ndarray):
        if np.isscalar(obj):
            return obj.dtype.type

        # TODO: deeper analysis
        return Array[obj.dtype, f"{obj.ndim}d"]

    if isinstance(obj, np.generic):
        return type(obj)

    raise NotImplementedError(
        f"Not able to determine the full type of {obj} (of type {type(obj)})"
    )


class ConstType(Type):
    """Private API class for const"""

    def __init__(self, type_):
        self.type = type_

    def format_as_backend_type(self, backend_type_formatter, **kwargs):
        return backend_type_formatter.make_const_code(
            format_type_as_backend_type(
                self.type, backend_type_formatter, **kwargs
            )
        )

    def __repr__(self):
        return f"const({repr(self.type)})"

    def is_fused_type(self):
        return self.type.is_fused_type()

    def get_template_parameters(self):
        return self.type.get_template_parameters()

    def short_repr(self):

        if hasattr(self.type, "short_repr"):
            short_repr_type = self.type.short_repr()
        else:
            short_repr_type = repr(self.type)

        return f"constI{short_repr_type}I"


def const(type_):
    """Declare a type as constant (``const`` C/Cython keyword)"""
    return ConstType(type_)



from transonic import boost, jit, set_backend_for_this_module, Transonic

set_backend_for_this_module("python")

ts = Transonic()


@boost
def func():
    return 0


@jit
def func_jit():
    return 0
